/**
  @file commandes.c
  Source contenant les definitions de toutes les commandes du langage KSI
*/

#include "commandes.h"
#include "systeme.h"

/**************************************************************************************
 Fait par Martin Ludovic
**************************************************************************************/

/**
 * Termine un processus proprement : 
 * avant du supprimer completement un processus (fonction supprimer_proc), il faut
 * - regarder qui est le pere de ce processus.
 *   - le pere est mort -> erreur !
 *   - ppid = NO_PERE -> On supprime completement ce processus \n
 *         (NO_PERE (0) est le "processus" qui tue tous les orphelins)
 *   - ppid != 0 -> On regarde s'il attend la mort d'un fils.
 *     - si oui :
 *       - ecrit ds P0 du pere le pid
 *       - lit $n ds P1 du pere et ecrit ds M$n du pere la valeur de sortie du processus
 *       - le reveille
 *     - si non : 
 *       - ecrit P0 (valeur de retour du processus) dans le mot NUM_MOT_EXIT de la table 
 *         des pages du processus
 *       - se met en E_ZOMBI
 * - regarder l'etat de ses fils:
 *   - E_ZOMBI -> les supprimer completement
 *   - autre -> les rendre orphelins
 * \param pid : Numero du processus qu'on veut supprimer
 * \return 
 * - -1 erreur: processus inchange
 * - -2 erreur: processus partiellement supprime (des pages sont probablement perdues)
 * - 0 sinon
 */
int terminer_proc (int pid) {
  mot_t mot_pid,mot_fils,mot_pere,mot_retour;
  int num_tab_p;
  int i;
  int erreur=0;
  int ppid;
  vad_t vad,vad2,vad3;

  if (lire_mot_mem(pid,NUM_PAGE_META,mot_pid) == -1)
    return -1;
  if (lire_etat(mot_pid) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Tentative de terminaison d'un proc mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot_pid)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_tab_p=rapatrier_page_swap(pid,(int)mot_pid[3])) < 0)
      return -1;
    break;
  case PP_RAM:
    num_tab_p = (int)mot_pid[3];
    break;
  }
  // Recherche ses fils
  for (i=1 ; i<N_MAX_PROC+1 ; i++) {
    switch (ppid=lire_ppid(i)) {
    case -1:
      if (lire_mot_mem(i,NUM_PAGE_META,mot_fils) == -1) {
        erreur = 1;
        break;
      }
      if (lire_etat(mot_fils) != E_MORT) 
        erreur = 1; 
      break;
    default:
      if (ppid == pid) {
        // Le proc i est un fils du processus a supprimer
        if (lire_mot_mem(i,NUM_PAGE_META,mot_fils) == -1) {
          erreur = 1;
          break;
        }
        switch (lire_etat(mot_fils)) {
        case E_ZOMBI:
          // Le fils est un zombi, on le supprime
          if (supprimer_proc(i) < 0)
            erreur = 1;
          break;
        default:
          // Le fils est processus actif, on le met orphelin
          if (changer_ppid(i,NO_PERE) == -1)
            erreur = 1;
          break;
        }
      }
    }
  }
  // Recherche son pere
  switch (ppid=lire_ppid(pid)) {
  case -1:
    erreur = 1;
    break;
  case NO_PERE:
    // Le processus a terminer est orphelin, on le supprime entierement
    if (supprimer_proc(pid) == -1)
      return -1;
    return 0;
  default:
    // Le proc pid n'est pas orphelin, son pere est ppid
    if (lire_mot_mem(ppid,NUM_PAGE_META,mot_pere) == -1) {
      erreur = 1;
      break;
    }
    switch (lire_etat(mot_pere)) {
    case E_MORT:
      // Le pere ne peut pas etre mort
      erreur = 1;
      // Mais on supprime qd meme le processus
      supprimer_proc(pid);
      break;
    case E_ATT_FILS:
      // Le pere attend la mort de son fils
      // Ecrit la pid ds P0 du pere
      vad2.num = 0;
      vad2.zone = Z_PILE;
      mot_retour[3] = (octet_t)pid;
      mot_retour[0] = (octet_t)M_UT;
      if (ecrire_mot(ppid,vad2,mot_retour) != 0)
        erreur = -1;
      // Lit P1 du pere
      vad2.num = 1;
      vad2.zone = Z_PILE;
      if (lire_mot(ppid,vad2,mot_retour) != 0)
        erreur = -1;
      // Ecrit ds M$n du pere la valeur de retour du proc
      vad2.num = (int)mot_retour[3];
      vad2.zone = Z_DONNEES;
      vad3.num = 0;
      vad3.zone = Z_PILE;
      if (lire_mot(pid,vad3,mot_retour) != 0)
        erreur = -1;
      if (ecrire_mot(ppid,vad2,mot_retour) != 0)
        erreur = -1;
      // On reveille le pere
      changer_etat(E_ELIGIBLE,mot_pere);
      if (ecrire_mot_mem(ppid,NUM_PAGE_META,mot_pere) == -1) {
        erreur = 1;
      }
      liste_add_pid(liste_proc,ppid);
      // On supprime completement son fils
      if (supprimer_proc(pid) < 0)
        erreur = 1;
      break;
    default:
      // Son pere n'attend pas sa mort
      // On sauvegarde la valeur de retour du proc
      vad.num = 0;
      vad.zone = Z_PILE;
      if (lire_mot(pid,vad,mot_retour) != 0)
        erreur = -1;
      if (ecrire_mot_mem(NUM_MOT_EXIT,num_tab_p,mot_retour) == -1)
        erreur = -1;
      // On marque le processus a terminer comme zombi
      changer_etat(E_ZOMBI,mot_pid);
      if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_pid) == -1) 
        erreur = 1;
      // On supprime toutes ses pages mais on garde une entree ds la table des proc
      if (supprimer_partie_proc(pid) < 0)
        erreur = 1;
    }
  }
  if (erreur == 0)
    return 0;
  return -2;
}

/**
 * Supprime un proc :
 * - libere son numero dans la table des processus
 * - libere toutes ses pages physiques (en memoire principale ou secondaire) 
 * \param pid : Numero du processus qu'on veut supprimer
 * \return 
 * - -1 erreur: processus inchange
 * - -2 erreur: processus partiellement supprime (des pages sont probablement perdues)
 * - 0 sinon
 */
int supprimer_proc(int pid) {
  mot_t mot_pid;
  int num_tab_p;
  int erreur = 0;

  // Cherche le num de la table des pages du processus pid
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_pid) == -1)
    return -1;
  if (lire_etat(mot_pid) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Tentative de suppression d'un proc mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot_pid)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_tab_p=rapatrier_page_swap(pid,(int)mot_pid[3])) < 0)
      return -1;
    break;
  case PP_RAM:
    num_tab_p = (int)mot_pid[3];
    break;
  }
  // Libere les pages attachees au processus
  if (supprimer_partie_proc(pid) < 0)
    erreur = 1;
  // Librere la page contenant la table des pages du processus
  if (liberer_page(num_tab_p) == -1)
    erreur = 1;
  // Libere le numero du processus
  if (liberer_pid(pid) == -1)
    erreur = 1;
  // Suppression de la liste des processus
  liste_supp_pid(liste_proc,pid);
  if (erreur == 0)
    return 0;
  else
    return -2;
}

/**
 * Supprime partiellement un processus :
 * - Supprime les pages physiques attachees au processus sauf la page qui contient 
 *   la table des pages car elle contient aussi 
 *   - le ppid (qui n'a pas pu etre mis dans la table des processus faute de place).
 *   - la valeur de retour du processus
 * - Supprime les signaux a intercepter de table_capture
 * \param pid : Numero du processus qu'on veut supprimer
 * \return 
 * - -1 erreur: processus inchange
 * - -2 erreur: processus partiellement supprime (des pages sont probablement perdues)
 * - 0 sinon
 */
int supprimer_partie_proc (int pid) {
  mot_t mot,mot_pid;
  int i,j;
  int num_tab_p;
  int num_mot;
  int erreur = 0;

  // Cherche le num de la table des pages du processus pid
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_pid) == -1)
    return -1;
  if (lire_etat(mot_pid) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Impossible de supprimer un proc mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot_pid)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_tab_p=rapatrier_page_swap(pid,(int)mot_pid[3])) < 0)
      return -1;
    break;
  case PP_RAM:
    num_tab_p = (int)mot_pid[3];
    break;
  }
  // Supprime les signaux captures
  supp_all_sig(pid);
  // Libere ses pages physiques
  for (i=0 ; i<3 ; i++) {
    for (j=i*T_MAX_ZONES ; j<(i+1)*T_MAX_ZONES ; j++) {
      if (lire_mot_mem(j,num_tab_p,mot) == -1)
        erreur = 1;
      else {
        if ((int)mot[2] == PP_SWAP) {
          // Supprime la page de la mem physique
          if (liberer_page_swap((int)mot[3]) != 0)
            erreur = 1;
          // Supprime l'entree ds la table des pages
          mot[2]= (octet_t)PP_NULL;
          if (ecrire_mot_mem(j,num_tab_p,mot) == -1)
            erreur = 1;
        }
        else if ((int)mot[2] == PP_RAM) {
          // Supprime la page de la mem physique
          if (liberer_page((int)mot[3]) != 0)
            erreur = 1;
          // Supprime l'entree ds la table des pages
          mot[2]= (octet_t)PP_NULL;
          if (ecrire_mot_mem(j,num_tab_p,mot) == -1)
            erreur = 1;
        }
        else if ((int)mot[2] == PP_NULL) {
          break;
        }
      }
    }
  }
  // Libere les pages tampon
  for (i=0 ; i<2; i++) {
    if (i == 0)
      num_mot = NUM_MOT_IN;
    else
      num_mot = NUM_MOT_OUT;
    if (lire_mot_mem(num_mot,num_tab_p,mot) == -1)
      erreur = 1;
    else {
      switch ((int)mot[2]) {
      case PP_SWAP:
        if (liberer_page_swap((int)mot[3]) != 0)
          erreur = 1;
        break;
      case PP_RAM:
        if (liberer_page((int)mot[3]) != 0)
          erreur = 1;
        break;
      }
    }
  }
  if (erreur == 0)
    return 0;
  else
    return -2;
}

/**
 * Regarde l'etat du drapeau du saut conditionnel.
 * L'information est stockee sur le 4eme bit de poids fort (00010000) du mot.
 * \param mot : mot correspondant a l'entree du processus ds la table des processus
 * \return 
 * - 1 si le drapeau est leve (ie la condition est vraie)
 * - 0 si le drapeau est baisse
 */
int condition_vraie (mot_t mot) {
  return ((mot[2] & 0x10) >> 4);
}

/**
 * Modifie l'etat du drapeau du saut conditionnel
 * L'information est stockee sur le 4eme bit de poids fort (00010000) du mot.
 * \param condition : le nouvel etat du drapeau
 * \param mot : mot correspondant a l'entree du processus ds la table des processus
 * \return l'ancien etat
 */
int changer_condition (int condition, mot_t mot) {
  int old;

  old = ((mot[2] & 0x10) >> 4);
  mot[2] = (mot[2] & 0xEF) | (condition << 4);
  return old;
}

/**
 * Parse un mot contenant une instruction. Extrait de ce mot, les 2 parmetres
 * de l'instruction ranges dans une structure.Cette fonction effectue egalement 
 * un pre-traitement en traitant le cas des indirections.
 * \param mot : mot contenant une instruction
 * \param tab_param : tableau ou sera stocke les 2 parametres
 * \param nb : nombre de parametres a parser
 * \return 
 * - -1 en cas d'erreur inconnue
 * - -2 en cas d'indirection erronnee
 * - 0 sinon
 */
int parser_param (int pid, mot_t mot, param_t tab_param[2], int nb) {
  int indirecte, ad, type, zone;
  vad_t ad_v;
  mot_t mot_ind;
  int i;

  for (i=0 ; i<nb ; i++) {
    if (i == 0) {
      // 1er parametre
      ad = (int)mot[2];
      zone = (int)((mot[1] & 0x10) >> 4) + 1;
      type = (int)((mot[1] & 0x20) >> 5);
      indirecte = (int)((mot[1] & 0x08) >> 3);
    }
    else {
      // 2eme parametre
      zone = (int)((mot[1] & 0x02) >> 1) + 1;
      ad = (int)mot[3];
      type = (int)((mot[1] & 0x04) >> 2);
      indirecte = (int)(mot[1] & 0x01);
    }
    // Traitement de l'indirection
    if (indirecte == P_INDIR) {
      ad_v.num = ad;
      ad_v.zone = zone;
      switch (lire_mot(pid,ad_v,mot_ind)) {
      case -1:
        return -1;
      case -2:
        return -2;
      default:
        ad = (int)mot_ind[3];
        zone = Z_DONNEES;
      }
    }
    tab_param[i].ad = ad;
    tab_param[i].type = type;
    tab_param[i].zone = zone;
  }
  return 0;
}

/**
 * Commande d'appel de fonction :
 * - Memorise le compteur programme sur la pile
 * - Modifie le compteur programme pour qu'il pointe sur la 1ere instruction de la fonction appelee
 * \param pid : num du processus en cours d'execution
 * \param mot : mot correspondant a l'instruction
 * \return
 * - -1 en cas d'echec
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_appel(int pid,mot_t mot) {
  int cp;
  mot_t mot_proc, mot_pile;
  vad_t ad_v;
  param_t tab_param[2];

  switch (parser_param(pid, mot, tab_param, 1)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Lit le compteur programme actuel
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  if (lire_etat(mot_proc) == E_MORT) {
    fprintf(stderr,"Tentative de modification d'un proc mort (%d)\n",pid);
    return -1;
  }
  cp = (int)mot_proc[1];
  // Memorise le compteur programme actuel
  if (changer_esp(pid,1) == -1)
    return -1;
  ad_v.num = 0;
  ad_v.zone = Z_PILE;
  mot_pile[0] = M_UT;
  mot_pile[3] = cp+1;
  switch (ecrire_mot(pid, ad_v, mot_pile)) {
  case -1:
    changer_esp(pid,-1);
    return -1;
  case -2:
    changer_esp(pid,-1);
    return -2;
  }
  // Modifie le compteur programme
  mot_proc[1] = (octet_t)tab_param[0].ad;
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    changer_esp(pid,-1);
    return -1;
  }
  return 0;
}

/**
 * Commande de retour de fonction
 * - Lit l'adresse de retour sur le haut de la pile
 * - Depile la pile de 1
 * - Modifie le compteur programme
 * \param pid : num du processus en cours d'execution
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sino
 */
int cmd_retour(int pid) {
  int cp;
  mot_t mot_proc, mot_pile;
  vad_t ad_v;

  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  if (lire_etat(mot_proc) == E_MORT) {
    fprintf(stderr,"Tentative de modification d'un proc mort (%d)\n",pid);
    return -1;
  }
  // Lit l'adresse de retour
  ad_v.num = 0;
  ad_v.zone = Z_PILE;
  switch (lire_mot(pid, ad_v, mot_pile)) {
  case -1:
    return -1;
  case -2:
    return -2;
  default:
    cp = (int)mot_pile[3];
  }
  // Modifie le compteur programme
  mot_proc[1] = (octet_t)cp;
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  // Depile la pile
  if (changer_esp(pid,-1) == -1)
    return -1;
  return 0;
}

/**
 * Commande pour fixer la taille de la zone de donnees.\n
 * Elle ne peut etre appelee qu'en debut de programme (compteur programme = 0)
 * \param pid : num du processus en cours d'execution
 * \param mot : mot correspondant a l'instruction
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 */
int cmd_donnees(int pid,mot_t mot) {
  mot_t mot_proc,mot_donnee, mot_tab;
  param_t tab_param[2];
  int cp;
  int taille;
  int num_page, num_tp;
  int i;

  if (parser_param(pid, mot, tab_param,1) == -1)
    return -1;
  // Verifie le type du 1er parametre
  if (tab_param[0].type != P_CONST) 
    return -1;
  // Cherche le compteur programme
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  if (lire_etat(mot_proc) == E_MORT) {
    fprintf(stderr,"Tentative de modification d'un proc mort (%d)\n",pid);
    return -1;
  }
  cp = (int)mot_proc[1];
  // DONNEES doit etre la 1ere instruction
  if (cp != 0)
    return -1;
  // Cherche la table des pages
  switch (lire_type_pp(mot_proc)) {
  case PP_RAM:
    num_tp = (int)mot_proc[3];
    break;
  case PP_SWAP:
    if ((num_tp=rapatrier_page_swap(pid,(int)mot_proc[3])) == -1)
      return -1;
    break;
  }
  // Initialise un mot de donnee
  mot_donnee[0] = (octet_t)M_UT;
  mot_donnee[3] = (octet_t)0;
  // Reserve la place necessaire aux donnees
  i = 0;
  taille = tab_param[0].ad;
  while (taille > 0) {
    // Rajoute une page a la zone de donnees
    if ((num_page=page_forcee(pid)) < 0)
      return -1;
    mot_tab[3] = (octet_t)num_page;
    mot_tab[2] = (octet_t)PP_RAM;
    if (ecrire_mot_mem(i+T_MAX_ZONES,num_tp,mot_tab) == -1)
      return -1;
    // Marque tous les mots de la zone de donnees comme utils
    for (i=0 ; (i<taille) && (i<NMPPP) ; i++) {
      if (ecrire_mot_mem(i,num_page,mot_donnee) == -1)
        return -1;
    }
    taille -= NMPPP;
    i++;
  }
  // Marque les mots en dehors de la zone de donnees comme inutilises
  mot_donnee[0] = (octet_t)M_NU;
  for (i=NMPPP+taille ; i<NMPPP ; i++) {
    if (ecrire_mot_mem(i,num_page,mot_donnee) == -1)
        return -1;
  }
  // Incremente le compteur programme
  cp ++;
  mot_proc[1] = (octet_t)cp;
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  return 0;
}

/**
 * Execute l'instruction AFFECTE ref = ref_cst.\n
 * Stocke a l'adresse du 1er arg la valeur du 2eme arg . Puis
 * deplace le compteur programme sur l'instruction suivante
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_affect (int pid, mot_t mot) {
  param_t tab_param[2];
  mot_t mot_val,mot_proc;
  vad_t adv;

  switch (parser_param(pid,mot, tab_param,2)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Verifie que le type des parametres est correct
  if (tab_param[0].type != P_REF)
    return -1;
  // Cherche la valeur du 2eme arg
  if (tab_param[1].type == P_CONST) {
    mot_val[3] = (octet_t)tab_param[1].ad;
    mot_val[2] = (octet_t)0;
    mot_val[1] = (octet_t)0;
    mot_val[0] = (octet_t)M_UT;
  }
  else {
    adv.num = tab_param[1].ad;
    adv.zone = tab_param[1].zone;
    switch (lire_mot(pid,adv,mot_val)) {
    case -1:
      return -1;
    case -2:
      return -2;
    }
  }
  // Ecrit cette valeur a l'adresse du 1er arg
  adv.num = tab_param[0].ad;
  adv.zone = tab_param[0].zone;
  switch (ecrire_mot(pid,adv,mot_val)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;  
  return 0;
}

/**
 * Execute l'instruction AFFECTE+ ref += ref_cst\n
 * Stocke a l'adresse du 1er arg la valeur du 2eme arg ajoutee a la valeur du 1er . Puis
 * deplace le compteur programme sur l'instruction suivante
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_affectp (int pid,mot_t mot) { 
  param_t tab_param[2];
  mot_t mot_val,mot_proc;
  vad_t adv;
  int val;

  switch (parser_param(pid,mot, tab_param, 2)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Verifie que le type des parametres est correct
  if (tab_param[0].type != P_REF)
    return -1;
  // Cherche la valeur du 1er arg
  adv.num = tab_param[0].ad;
  adv.zone = tab_param[0].zone;
  switch (lire_mot(pid,adv,mot_val)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  val = (int)mot_val[3];
  // Cherche la valeur du 2eme arg
  if (tab_param[1].type == P_CONST) {
    mot_val[3] = (octet_t)(tab_param[1].ad + val);
    mot_val[0] = (octet_t)M_UT;
  }
  else {
    adv.num = tab_param[1].ad;
    adv.zone = tab_param[1].zone;
    switch (lire_mot(pid,adv,mot_val)) {
    case -1:
      return -1;
    case -2:
      return -2;
    }
    mot_val[3] = (octet_t)((int)mot_val[3] + val);
  }
  // Ecrit cette valeur a l'adresse du 1er arg
  adv.num = tab_param[0].ad;
  adv.zone = tab_param[0].zone;
  switch (ecrire_mot(pid,adv,mot_val)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;  
  return 0;
}

/**
 * Execute l'instruction AFFECTESP
 * Stocke a l'adresse du 1er arg la valeur du 2eme arg qui correspond a un pointeur
 * sur fonction.Puis deplace le compteur programme sur l'instruction suivante.
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_affectsp (int pid, mot_t mot) {
  param_t tab_param[2];
  mot_t mot_val,mot_proc;
  vad_t adv;

  switch (parser_param(pid,mot, tab_param,2)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Cherche la valeur du 2eme arg
  mot_val[3] = (octet_t)tab_param[1].ad;
  mot_val[0] = (octet_t)M_UT;
  // Ecrit cette valeur a l'adresse du 1er arg
  adv.num = tab_param[0].ad;
  adv.zone = tab_param[0].zone;
  switch (ecrire_mot(pid,adv,mot_val)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) 
    return -1;  
  return 0;
}

/**
 * Execute l'instruction TEST.\n
 * Cette commande verifie l'egalite (en valeur) des 2 parametres puis leve ou abaisse
 * en consequence le drapeau pour le saut conditionnnel. Elle incremente aussi le 
 * compteur programme.
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_test(int pid,mot_t mot) {
  param_t tab_param[2];
  int val1,val2;
  mot_t mot_val,mot_proc;
  vad_t adv;

  switch (parser_param(pid,mot, tab_param, 2)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Cherche la valeur du 1er arg
  if (tab_param[0].type == P_CONST) {
    val1 = tab_param[0].ad;
  }
  else {
    adv.num = tab_param[0].ad;
    adv.zone = tab_param[0].zone;
    switch (lire_mot(pid,adv,mot_val)) {
    case -1:
      return -1;
    case -2:
      return -2;
    }
    val1 = (int)mot_val[3];
  }
  // Cherche la valeur du 2eme arg
  if (tab_param[1].type == P_CONST) {
    val2 = tab_param[1].ad;
  }
  else {
    adv.num = tab_param[1].ad;
    adv.zone = tab_param[1].zone;
    switch (lire_mot(pid,adv,mot_val)) {
    case -1:
      return -1;
    case -2:
      return -2;
    }
    val2 = (int)mot_val[3];
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  // Compare les 2 valeurs
  if (val2 == val1)
    changer_condition(C_VRAI,mot_proc);
  else
    changer_condition(C_FAUX,mot_proc);
  //  MAJ de la table des processus
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return -1;
  return 0;
}

/**
 * Execute l'instruction SI :
 * - si le drapeau des sauts conditionnel est leve -> effectue le saut
 * - sinon deplace juste le compteur programme a l'instruction suivante
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_si (int pid,mot_t mot){
  param_t tab_param[2];
  mot_t mot_proc;

  switch (parser_param(pid,mot, tab_param, 1)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Lecture de l'entree du proc ds la table des processus
  if (lire_mot_mem(pid, NUM_PAGE_META,mot_proc) == -1)
    return -1;
  // Calcule du nouveau compteur programme
  if (condition_vraie(mot_proc)) 
    mot_proc[1] = (octet_t)tab_param[0].ad;
  else 
    mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  // MAJ de la table des proc
  if (ecrire_mot_mem(pid, NUM_PAGE_META, mot_proc) == -1)
    return -1;
  return 0;
}

/**
 * Execute l'instruction SAUT.\n
 * Deplace le compteur programme a l'endroit indique par le 1er parametre.
 * \param pid : pid du processus
 * \param mot : mot correspondant  l'instruction
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_saut (int pid,mot_t mot) {
  param_t tab_param[2];
  mot_t mot_proc;

  switch (parser_param(pid,mot, tab_param, 1)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Lecture de l'entree du proc ds la table des processus
  if (lire_mot_mem(pid, NUM_PAGE_META,mot_proc) == -1)
    return -1;
  mot_proc[1] = (octet_t)tab_param[0].ad;
  // MAJ de la table des proc
  if (ecrire_mot_mem(pid, NUM_PAGE_META, mot_proc) == -1)
    return -1;
  return 0;
}

/**
 * Commande TRAPPE qui retourne le pid du processus. C'est a dire qui l'ecrit sur la pile\n
 * Incremente le compteur programme.
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_id (int pid) {
  vad_t adv;
  mot_t mot,mot_proc;

  // Ecrit le pid en haut de la pile
  mot[0] = (octet_t)M_UT;
  mot[3] = (octet_t)pid;
  adv.num = 0;
  adv.zone = Z_PILE;
  switch (ecrire_mot(pid,adv,mot)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    mot[3] = (int)0;
    ecrire_mot(pid,adv,mot);
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    mot[3] = (int)0;
    ecrire_mot(pid,adv,mot);
    return -1;  
  }
  return 0;
}

/**
 * Commande TRAPPE qui ecrit sur le haut de la pile le ppid du processus.\n
 * Incremente le compteur programme.
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_idp(int pid){
  vad_t adv;
  mot_t mot,mot_proc;
  int ppid;

  // Cherche le ppid du processus
  if ((ppid=lire_ppid(pid)) == -1)
    return -1;
  // Ecrit le ppid en haut de la pile
  mot[0] = (octet_t)M_UT;
  mot[3] = (octet_t)ppid;
  adv.num = 0;
  adv.zone = Z_PILE;
  switch (ecrire_mot(pid,adv,mot)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    mot[3] = (int)0;
    ecrire_mot(pid,adv,mot);
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    mot[3] = (int)0;
    ecrire_mot(pid,adv,mot);
    return -1;  
  }
  return 0;
}

/**
 * Commande TRAPPE qui associe un handler a un signal pour le processus.\n
 * Incremente le compteur programme.\n
 * Ses arguments:
 * - P1: le signal
 * - P2: pointeur sur le handler
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_capture (int pid) {
  vad_t adv, adv_err;
  mot_t mot_arg, mot_err, mot_proc;
  signal_t *sig;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Cherche le signal a capturer (dans P1)
  adv.num = 1;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_arg)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  sig = malloc(sizeof(signal_t));
  if ((sig->signal=sig_ksinu2lib((int)mot_arg[3])) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    free(sig);
    return -1;
  }
  // Associe a ce signal son nouveau comportement (dans P2)
  adv.num = 2;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_arg)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    free(sig);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    free(sig);
    return -2;
  }
  sig->cp = (int)mot_arg[3];
  add_sig(pid,sig);
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;  
  } 
  // Retourne 0 en cas de reussite
  mot_err[3] = (int)0;
  ecrire_mot(pid,adv_err,mot_err);
  return 0;
}

/**
 * Commande TRAPPE qui envoie a un processus un signal\n
 * Incremente le compteur programme.\n
 * Ses arguments:
 * - P1: le numero de processus
 * - P2: le signal
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_emets (int pid) {
  vad_t adv,adv_err;
  mot_t mot_arg, mot_err, mot_proc;
  int pid2;
  int sig;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lecture du num de processus auquel envoyer le signal
  adv.zone = Z_PILE;
  adv.num = 1;
  switch (lire_mot(pid,adv,mot_arg)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  pid2 = (int)mot_arg[3];
  // Lecture du signal a envoyer
  adv.num = 2;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_arg)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  if ((sig=sig_ksinu2lib((int)mot_arg[3])) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Envoie du signal
  if (lire_mot_mem(pid2,NUM_PAGE_META, mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  mot_proc[0] = mot_proc[0] | sig;
  if (ecrire_mot_mem(pid2,NUM_PAGE_META, mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Deplace le compteur programme sur l'instruction suivante
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;  
  } 
  // Retourne 0 en cas de reussite
  mot_err[3] = (int)0;
  ecrire_mot(pid,adv_err,mot_err);
  return 0;
}

/**
 * Commande TRAPPE qui lit une lettre ds le tube d'entree. La lecture est bloquante.\n
 * Incremente le compteur programme.\n
 * Ses arguments:
 * - P1: l'adresse (Mx) ou on ecrit le caractere
 * - P2: le nombre de carateres a lire (bloque ici a 1)
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_lit (int pid){
  char c;
  mot_t mot_proc, mot_err,mot_pile;
  vad_t adv_err,adv;
  int ad,nb;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lit l'entree du proc ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Lecture du 1er arg (P1)
  adv.num = 1;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  ad = (int)mot_pile[3];
  // Lecture du 2eme arg (P2)
  adv.num = 2;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  nb = 1; // normalement nb = (int)mot_pile[3];
  // for (i=0 ; i<nb ; i++) ...
  // Lit un caractere du tampon d'entree
  switch (lire_stdin(pid,&c)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
  case -3:
    // Mettre le processus en attente de lecture
    changer_etat(E_ATT_LECT,mot_proc);
    if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
      ecrire_mot(pid,adv_err,mot_err);
      return -1;
    }
    liste_supp_pid(liste_proc,pid);
    return 0;
  default:
    // Lecture reussie -> on ecrit le caract lu ds M ad
    adv.num = ad;
    adv.zone = Z_DONNEES;
    mot_pile[0] = (octet_t)M_UT;
    mot_pile[3] = (octet_t)c;
    switch (ecrire_mot(pid,adv,mot_pile)) {
    case -1:
      ecrire_mot(pid,adv_err,mot_err);
      return -1;
    case -2:
      ecrire_mot(pid,adv_err,mot_err);
      return -2;
    }
  }
  // Deplace le compteur programme sur l'instruction suivante
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;  
  } 
  // Retourne 0 ds P0
  adv.num = 0;
  adv.zone = Z_PILE;
  mot_pile[0] = (octet_t)M_UT;
  mot_pile[3] = (octet_t)0;
  switch (ecrire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  return 0;
}

/**
 * Commande TRAPPE qui ecrit une lettre ds le tube de sortie. Commande bloquante\n
 * Incremente le compteur programme.\n
 * Ses arguments:
 * - P1: l'adresse (Mx) ou est stocke le caractere a ecrire
 * - P2: le nombre de carateres a ecrire (bloque ici a 1)
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_ecrit(int pid){
  char c;
  mot_t mot_proc, mot_err,mot_pile;
  vad_t adv_err,adv;
  int ad,nb;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lit l'entree du proc ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Lecture du 1er arg (P1)
  adv.num = 1;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  ad = (int)mot_pile[3];
  // Lecture du caract a ecrire
  adv.num = ad;
  adv.zone = Z_DONNEES;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  c = (char)mot_pile[3];
  // Lecture du 2eme arg (P2)
  adv.num = 2;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  nb = 1; // normalement nb = (int)mot_pile[3];
  // for (i=0 ; i<nb ; i++) ...
  // Lit un caractere du tampon d'entree
  switch (ecrire_stdout(pid,c)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
  case -3:
    // Mettre le processus en attente de lecture
    changer_etat(E_ATT_ECRIT,mot_proc);
    if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
      ecrire_mot(pid,adv_err,mot_err);
      return -1;
    }
    liste_supp_pid(liste_proc,pid);
    return 0;
  }
  // Deplace le compteur programme sur l'instruction suivante
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;  
  } 
  // Ecriture reussie -> retourne 0 ds P0
  adv.num = 0;
  adv.zone = Z_PILE;
  mot_pile[0] = (octet_t)M_UT;
  mot_pile[3] = (octet_t)0;
  switch (ecrire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  return 0;
}

/**
 * Commande TRAPPE terminant le processus\n
 * Incremente le compteur programme.\n
 * Ses arguments:
 * - P0: la valeur de retour du processus
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_fin(int pid){
  printf("TRAPPE FIN proc %d\n",pid);
  if (terminer_proc(pid) < 0)
    return -1;
  return 0;
}

/**
 * Commande TRAPPE qui recouvre l'espace memoire d'un processus pour executer un autre programme.\n
 * Incremente le compteur programme.\n
 * Argument : l'adresse ou est stocke le nom du fichier objet (1 caractere)
 * Retourne : 
 * - -1 en cas d'erreur
 * - rien sinon
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_recouvre (int pid) {
  vad_t adv_err,adv;
  mot_t mot_pile,mot_err,mot_proc;
  char *nom;
  int ppid;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lit l'entree du proc ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  if (lire_etat(mot_proc) == E_MORT) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Lit l'argument
  adv.num = 1;
  adv.zone = Z_PILE;
  switch (lire_mot(pid,adv,mot_pile)) {
  case -1:
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  case -2:
    ecrire_mot(pid,adv_err,mot_err);
    return -2;
  }
  nom = malloc(8);
  sprintf(nom,"%c.objet",(char)mot_pile[3]);
  // Supprime partiellemnt le processus
  if (supprimer_partie_proc(pid) < 0) {
    ecrire_mot(pid,adv_err,mot_err);
    free(nom);
    return -1; 
  }
  // Sauvegarde le pid du pere
  if ((ppid=lire_ppid(pid)) == -1) {
    free(nom);
    return -1;
  }
  // Supprime la table des pages
  switch (lire_type_pp(mot_proc)) {
  case PP_SWAP:
    if (liberer_page_swap((int)mot_proc[3]) == -1) {
      free(nom);
      return -1;
    }
  case PP_RAM:
    if (liberer_page((int)mot_proc[3]) == -1) {
      free(nom);
      return -1;
    }
  }
  // Lance le nouveau programme
  if (lancer_prog(nom,pid) == -1) {
    free(nom);
    return -1;
  }
  // Indique son ppid
  if (changer_ppid(pid,ppid) == -1) {
    free(nom);
    return -1;
  }
  free(nom);
  return 0;
}

/**
 * Commande TRAPPE qui duplique l'espace memoire\n
 * Incremente le compteur programme.\n
 * Aucun argument.\n
 * Retourne :
 * - 0 ds l'espace memoire du fils 
 * - le pid du fils si on est ds l'espace memoire du pere
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_clone (int pid) {
  int ppid_f = pid;
  int pid_f;
  int i,j;
  mot_t mot_proc, mot_proc_f, mot_tp_f, mot_tp, mot_pp, mot_err;
  vad_t adv_err;
  int num_tp, num_tp_f, num_pp_f, num_pp;

  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lit l'entree du proc ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  if (lire_etat(mot_proc) == E_MORT) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Lecture de la table des pages
  switch (lire_type_pp(mot_proc)) {
  case PP_RAM:
    num_tp = (int)mot_proc[3];
    liste_acces_page(liste_pages,num_tp);
    break;
  case PP_SWAP:
    if ((num_tp=rapatrier_page_swap(pid,(int)mot_proc[3])) < 0) {
      ecrire_mot(pid,adv_err,mot_err);
      return -1;
    }
    break;
  }
  // Obtient un nouveau pid
  if ((pid_f=pid_libre()) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Obtient une nouvelle table des pages
  if ((num_tp_f=page_forcee(pid_f)) < 0) {
    liberer_pid(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Incremente le compteur programme
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1]+1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Initialisation de l'entree ds la table des proc
  mot_proc_f[0] = mot_proc[0];
  mot_proc_f[1] = mot_proc[1];
  mot_proc_f[2] = mot_proc[2];
  changer_type_pp(PP_RAM,mot_proc_f);
  mot_proc_f[3] = num_tp_f;
  if (ecrire_mot_mem(pid_f,NUM_PAGE_META,mot_proc_f) == -1) {
    liberer_pid(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Remplissage de la table des pages du fils
  for (i=0 ; i<3*T_MAX_ZONES+3 ; i++) {
    if (i != NUM_MOT_IN+1) { // NUM_MOT_IN+1 n'est pas un mot pointant sur une page
      // Obtient une page RAM libre
      if ((num_pp_f=page_forcee(pid_f)) == -1) {
        supprimer_proc(pid_f);
        ecrire_mot(pid,adv_err,mot_err);
        return -1;
      }
      // Lecture du mot de la table du pere
      if (lire_mot_mem(i, num_tp, mot_tp) == -1) {
        supprimer_proc(pid_f);
        ecrire_mot(pid,adv_err,mot_err);
        liberer_page(num_pp_f);
        return -1;
      }
      switch ((int)mot_tp[2]) {
      case PP_NULL:
        // Recopie du mot
        mot_tp_f[2] = (octet_t)PP_NULL;
        mot_tp_f[3] = (octet_t)0;
        if (ecrire_mot_mem(i,num_tp_f,mot_tp_f) == -1) {
          supprimer_proc(pid_f);
          ecrire_mot(pid,adv_err,mot_err);
          liberer_page(num_pp_f);
          return -1;
        }
        // mais pas de la page
        liberer_page(num_pp_f);
        num_pp_f = -1;
        break;
      case PP_RAM:
        num_pp = (int)mot_tp[3];
        break;
      case PP_SWAP:
        if ((num_pp=rapatrier_page_swap(pid,(int)mot_tp[3])) == -1) {
          supprimer_proc(pid_f);
          liberer_page(num_pp_f);
          ecrire_mot(pid,adv_err,mot_err);
          return -1;
        }
        break;
      }
      if (num_pp_f != -1) {
        // Recopie de la page mot par mot
        for (j=0 ; j<NMPPP ; j++) {
          if (lire_mot_mem(j,num_pp,mot_pp) == -1) {
            supprimer_proc(pid_f);
            ecrire_mot(pid,adv_err,mot_err);
            liberer_page(num_pp_f);
            return -1;
          }
          if (ecrire_mot_mem(j,num_pp_f,mot_pp) == -1) {
            supprimer_proc(pid_f);
            liberer_page(num_pp_f);
            ecrire_mot(pid,adv_err,mot_err);
            return -1;
          }
        }
        // Inscrit cette page ds la table des pages du fils
        mot_tp_f[3] = (octet_t)num_pp_f;
        mot_tp_f[2] = (octet_t)PP_RAM;
        mot_tp_f[0] = mot_tp[0];
        mot_tp_f[1] = mot_tp[1];
        if (ecrire_mot_mem(i, num_tp_f, mot_tp_f) == -1) {
          supprimer_proc(pid_f);
          liberer_page(num_pp_f);
          ecrire_mot(pid,adv_err,mot_err);
          return -1;
        }
      }
    }
  }
  // Initialise la derniere zone de la table des pages
  // MAJ la taille de la pile
  if (lire_mot_mem(NUM_MOT_ESP,num_tp,mot_tp) == -1) {
    supprimer_proc(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  if (ecrire_mot_mem(NUM_MOT_ESP,num_tp_f,mot_tp) == -1) {
    supprimer_proc(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Deverrouillage de tampons
  mot_tp_f[3] = (octet_t)F_UNLOCK;
  if (ecrire_mot_mem(NUM_MOT_IN+1,num_tp_f,mot_tp_f) == -1) {
    supprimer_proc(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  if (ecrire_mot_mem(NUM_MOT_OUT+1,num_tp_f,mot_tp_f) == -1) {
    supprimer_proc(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Indiquer ppid
  mot_tp_f[3] = (octet_t)ppid_f;
  if (ecrire_mot_mem(NUM_MOT_PPID,num_tp_f,mot_tp_f) == -1) {
    supprimer_proc(pid_f);
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Retourne le pid du fils en haut de la pile du pere
  mot_err[3] = (octet_t)pid_f;
  ecrire_mot(pid,adv_err,mot_err);
  // Retourne 0 en haut de la pile du fils
  mot_err[3] = (octet_t)0;
  ecrire_mot(pid_f,adv_err,mot_err);
  return 0;
}

/**
 * Commande TRAPPE qui suspend le processus jusqu' la mort d'un fils.\n
 * Incremente le compteur programme.\n
 * Un argument : P1 l'adresse memoire ou stocker la valeur de retour du fils
 * Retourne :
 * - 255 en cas d'echec
 * - le pid du fils sinon
 * \param pid : pid du processus
 * \return
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_attends(int pid) {
  int i;
  mot_t mot_proc, mot_err, mot_fils, mot_retour, mot_ad;
  vad_t adv_err,adv_ad;
  int num_tp_f;
  int ppid;

  printf("TRAPPE ATTENDS proc %d\n",pid);
  // Retourne 255 en cas d'erreur
  adv_err.num = 0;
  adv_err.zone = Z_PILE;
  mot_err[0] = (octet_t)M_UT;
  mot_err[3] = (octet_t)255;
  // Lit l'entree du proc ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  if (lire_etat(mot_proc) == E_MORT) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Incremente le compteur programme
  mot_proc[1] = (octet_t)((int)mot_proc[1]+1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  // Recherche ses fils
  for (i=1 ; i<N_MAX_PROC+1 ; i++) {
    if (i != pid) {
      switch (ppid=lire_ppid(i)) {
      case -1:
        if (lire_mot_mem(i,NUM_PAGE_META,mot_fils) == -1) {
          ecrire_mot(pid,adv_err,mot_err);
          return -1;
        }
        if (lire_etat(mot_fils) != E_MORT) {
          ecrire_mot(pid,adv_err,mot_err);
          return -1;
        }
        break;
      default:
        if (ppid == pid) {
          // Le proc i est un fils du processus a supprimer
          if (lire_mot_mem(i,NUM_PAGE_META,mot_fils) == -1) {
            ecrire_mot(pid,adv_err,mot_err);
            return -1;
          }
          if (lire_etat(mot_fils) == E_ZOMBI) { // Le fils est un zombi
            // Lecture de sa table des pages (pour la valeur de retour)
            switch (lire_type_pp(mot_fils)) {
            case PP_RAM:
              num_tp_f = (int)mot_fils[3];
              break;
            case PP_SWAP:
              if ((num_tp_f=rapatrier_page_swap(i,(int)mot_fils[3])) == -1) {
                ecrire_mot(pid,adv_err,mot_err);
                return -1;
              }
              break;
            }
            // Recupere la valeur de retour du fils
            if (lire_mot_mem(NUM_MOT_EXIT,num_tp_f,mot_retour) == -1) {
              ecrire_mot(pid,adv_err,mot_err);
              return -1;
            }
            // On lit l'argument de la commande
            adv_ad.num = 1;
            adv_ad.zone = Z_PILE;
            switch (lire_mot(pid,adv_ad,mot_ad)) {
            case -1:
              ecrire_mot(pid,adv_err,mot_err);
              return -1;
            case -2:
              ecrire_mot(pid,adv_err,mot_err);
              return -2;
            }
            // On ecrit la valeur de retour a l'adresse donnee par l'argument
            adv_ad.num = (int)mot_ad[3];
            adv_ad.zone = Z_DONNEES;
            mot_retour[0] = (octet_t)M_UT;
            switch (ecrire_mot(pid,adv_ad,mot_retour)) {
            case -1:
              ecrire_mot(pid,adv_err,mot_err);
              return -1;
            case -2:
              ecrire_mot(pid,adv_err,mot_err);
              return -2;
            }
            // On retourne le pid du fils
            adv_ad.num = 0;
            adv_ad.zone = Z_PILE;
            mot_retour[3] = (octet_t)i;
            switch (ecrire_mot(pid,adv_ad,mot_retour)) {
            case -1:
              ecrire_mot(pid,adv_err,mot_err);
              return -1;
            case -2:
              ecrire_mot(pid,adv_err,mot_err);
              return -2;
            }
            // On supprime le fils
            if (supprimer_proc(i) < 0) {
              ecrire_mot(pid,adv_err,mot_err);
              return -1;
            }
            return 0;
          }
        }
      }
    }
  }
  // Aucun fils n'est deja mort
  // Le processus change d'etat
  changer_etat(E_ATT_FILS,mot_proc);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    ecrire_mot(pid,adv_err,mot_err);
    return -1;
  }
  liste_supp_pid(liste_proc,pid);
  return 0;
}



/**************************************************************************************
 Fait par Ouachani Mourad
**************************************************************************************/



/**
 * Augmente la taille de la pile.\n
 * Et incremente la valeur du compteur programme
 * \param pid : numero du processus dont on veut augementer la taille de la pile
 * \param nbre : nombre de mot a augementer 
 * \return 
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_cpile(int pid,mot_t mot){
  int i;
  int op = 1;
  param_t tab_param[2];
  int nbre;
  mot_t mot_proc;

  switch (parser_param(pid, mot, tab_param, 1)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Verifie que le type des parametres est correct
  if (tab_param[0].type != P_CONST)
    return -1;
  nbre = tab_param[0].ad;
  if(nbre < 1){
    printf("Verifier le nombre de mot a augementer dans la pile");
    return -1;
  }
  for(i=0; i< nbre;i++){
    if(changer_esp(pid,op) == -1){
      printf("erreur : impossible d'incrementer la hauteur de la pile\n");
      return -1;
    }
  }
  // Incremente le compteur programme
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    return -1;  
  }
  return 0;
}

/**
 * Diminue la taille de la pile.\n
 * En plus de modifier la taille :
 * - incremente la valeur du compteur programme\n
 * \param pid : numero du processus dont on veut diminuer la taille de la pile
 * \param nbre : nombre de mot a diminuer 
 * \return 
 * - -1 en cas d'erreur
 * - -2 en cas de violation de memoire
 * - 0 sinon
 */
int cmd_dpile(int pid,mot_t mot){
  int i;
  int op = -1;
  param_t tab_param[2];
  int nbre;
  mot_t mot_proc;

  switch (parser_param(pid,mot, tab_param, 2)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Verifie que le type des parametres est correct
  if (tab_param[0].type != P_CONST)
    return -1;
  nbre = tab_param[0].ad;
  if(nbre < 1){
    printf("Verifier le nombre de mot a diminuer dans la pile");
    return -1;
  }
  for(i=0; i< nbre;i++){
    if(changer_esp(pid,op) == -1){
      printf("erreur : impossible de decrementer la hauteur de la pile\n");
      return -1;
    }
  }
  // Incremente le compteur programme
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    return -1;
  }
  mot_proc[1] = (octet_t)((int)mot_proc[1] + 1);
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    return -1;  
  }
  return 0;
}
