/**
 * \file fifo.c
 * Source definissant toutes les fonctions utiles pour manipuler une pile FIFO.
 * Et notament pour gerer la liste de processus actifs et l'ordonnancement de leur execution.
 */
#include "fifo.h"

/**************************************************************************************
 Entierement fait par Martin Ludovic
**************************************************************************************/

/**
 * Initialise une pile
 * \param liste : la pile a initialiser
 * \return rien
 */
void liste_init (l_fifo* liste) {
  liste->tete = NULL;
}

/**
 * Ajoute un nouveau processus a la pile
 * \param liste : la pile ds laquelle il faut ajouter un processus
 * \param pid : le num du processus a ajouter
 * \return rien
 */
void liste_add_pid (l_fifo* liste, int pid) {
  struct cellule *p_cel; 
  struct cellule *p_tmp; 

  // Initialisation de la cellule
  p_cel = malloc(sizeof(struct cellule));
  p_cel->pid = pid;
  p_cel->temps = 0;
  p_cel->suivant = NULL;
  // Cas de la liste vide
  if (liste->tete == NULL) {
    liste->tete = p_cel;
  }
  else { // Liste non vide
    p_tmp = liste->tete;
    while (p_tmp->suivant) {
      p_tmp = p_tmp->suivant;
    }
    p_tmp->suivant = p_cel;
  }
}

/**
 * Retourne le num du processus dont il faut executer une instruction
 * \param liste : la pile contenant la liste des proc en cour d'exec
 * \return
 * - -1 en cas de pile vide
 * - le num du processus sinon
 */
int liste_next_pid (l_fifo* liste){
  struct cellule *p_tmp; 

  // Cas de la liste vide
  if (liste->tete == NULL) {
    fprintf(stderr,"Impossible de determiner le num du prochain processus : la liste est vide\n");
    return -1;
  }
  else{
    // Un processus passe la main
    if (liste->tete->temps >= quantum) {
      liste->tete->temps = 0;
      p_tmp = liste->tete;
      while (p_tmp->suivant) {
        p_tmp = p_tmp->suivant;
      }
      p_tmp->suivant = liste->tete;
      liste->tete = liste->tete->suivant;
      p_tmp->suivant->suivant = NULL;
    }
    liste->tete->temps ++;
    return liste->tete->pid;
  }
}

/**
 * Supprime un processus de la pile
 * \param liste : la pile contenant la liste des proc en cour d'exec
 * \param pid : le num du processus a supprimer
 * \return 
 * - -1 en cas de pile vide
 * - -2 si il n'y a pas de processus de num pid
 * -  0 sinon
 */
int liste_supp_pid (l_fifo* liste, int pid) {
  struct cellule *p_tmp; 
  struct cellule *p_clr; 

  // Cas de la liste vide
  if (liste->tete == NULL) {
    fprintf(stderr,"Impossible de supprimer un element d'une liste vide\n");
    return -1;
  }
  else{
    if (liste->tete->pid == pid) {
      p_clr = liste->tete;
      liste->tete =  liste->tete->suivant;
      free (p_clr);
      return 0;
    }
    p_tmp = liste->tete;
    while (p_tmp->suivant) {
      if (p_tmp->suivant->pid == pid) {
        p_clr = p_tmp->suivant;
        p_tmp->suivant = p_tmp->suivant->suivant;
        free(p_clr);
        return 0;
      }
      p_tmp = p_tmp->suivant;
    }
    return -2; // Le processus n'a pas ete trouve
  }
}    

/**
 * Ajoute une nouvelle page a la pile.
 * \param liste : la pile ds laquelle il faut ajouter une page
 * \param page : le num de la page a ajouter
 * \return rien
 */
void liste_add_page (l_fifo* liste, int page) {
  struct cellule *p_cel; 
  struct cellule *p_tmp; 

  // Initialisation de la cellule
  p_cel = malloc(sizeof(struct cellule));
  p_cel->page = page;
  p_cel->suivant = NULL;
  // Cas de la liste vide
  if (liste->tete == NULL) {
    liste->tete = p_cel;
  }
  else { // Liste non vide
    p_tmp = liste->tete;
    while (p_tmp->suivant) {
      p_tmp = p_tmp->suivant;
    }
    p_tmp->suivant = p_cel;
  }
}

/**
 * Retourne le num de la plus vielle page
 * \param liste : la pile contenant la liste des pages physiques de la  memoire principale
 * \return
 * - -1 en cas d'erreur
 * - Le num de la plus vielle page
 */
int liste_old_page (l_fifo* liste) {
  // Cas de la liste vide
  if (liste->tete == NULL) {
    fprintf(stderr,"Impossible de determiner le num de la plus vielle page : la liste est vide\n");
    return -1;
  }
  else // Liste non vide
    return liste->tete->page;
}

/**
 * Quand on accede a une page, elle devient la page la plus 'jeune'. Cette fonction permet 
 * d'actualiser l'age de cette page. En fait, on deplace cette page au fond de la pile.
 * \param liste : la pile contenant la liste des pages physiques de la  memoire principale
 * \param page : le num de la page a deplacer
 * \return 
 * - -1 en cas de pile vide
 * - -2 si il n'y a pas de page de num page
 * -  0 sinon
 */
int liste_acces_page (l_fifo* liste, int page) {
  switch (liste_supp_page(liste,page)) {
  case -1:
    return -1;
  case -2:
    return -2;
  default:
    // La page 'page' a ete trouvee et supprimee
    // Il faut la rajouter a la fin de la liste
    liste_add_page(liste,page);
    return 0;
  }
}

/**
 * Supprime une page de la pile.
 * \param liste : la pile contenant la liste des pages physiques de la  memoire principale
 * \param page : le num de la page a supprimer
 * \return 
 * - -1 en cas de pile vide
 * - -2 si il n'y a pas de page de num page
 * -  0 sinon
 */
int liste_supp_page (l_fifo* liste, int page) {
  struct cellule *p_tmp; 
  struct cellule *p_clr; 

  if (liste->tete == NULL) {  // Cas de la liste vide
    fprintf(stderr,"Impossible de supprimer un element d'une liste vide\n");
    return -1;
  }
  if (liste->tete->page == page) {
    p_clr = liste->tete;
    liste->tete =  liste->tete->suivant;
    free (p_clr);
    return 0;
  }
  p_tmp = liste->tete;
  while (p_tmp->suivant) {
    if (p_tmp->suivant->page == page) {
      p_clr = p_tmp->suivant;
      p_tmp->suivant = p_tmp->suivant->suivant;
      free(p_clr);
      return 0;
    }
    p_tmp = p_tmp->suivant;
  }
  return -2; // La page n'a pas ete trouvee
}

/**
 * Retourne la taille de la liste
 * \param liste : la pile contenant la liste des proc en cour d'execution
 * \return 
 * La taille de la liste
 */
int liste_taille (l_fifo* liste) {
  struct cellule *p_tmp; 
  int taille;

  taille = 0;
  p_tmp = liste->tete;
  while (p_tmp != NULL) {
    taille ++;
    p_tmp = p_tmp->suivant;
  }
  return taille;
}

/**
 * Retourne un tableau d'entiers representant la liste des processus en cours d'utilisation
 * \param liste : la pile contenant la liste des proc en cour d'execution
 * \return 
 * Un pointeur sur un tableau d'entiers representant la liste passee en parametre.
 */
int *liste_enum_pid (l_fifo* liste) {
  struct cellule *p_tmp; 
  int * tab;
  int taille;
  int i;

  taille = liste_taille(liste);
  tab = malloc(taille * sizeof(int));
  p_tmp = liste->tete;
  for (i=0 ; i<taille ; i++) {
    tab[i] = p_tmp->pid;
    p_tmp = p_tmp->suivant;
  }
  return tab;
}
