/**
 * @file interpreteur.c
 * Source definissant toutes les fonctions utiles pour interpreter une instruction
 * et par extention toutes les fonctions pour manipuler les processus et la mmoire.
 */

#include "interpreteur.h"
#include "commandes.h"

/**************************************************************************************
 Entierement fait par Martin Ludovic
**************************************************************************************/

/**
 * Fonction qui permet de changer l'etat d'un processus
 * L'information est stockee sur les bits 2, 3 et 4 par ordre de poids croissant
 * (00001110) du 2eme octet de poids faible (ie le 3eme en partant de la droite) 
 * du mot
 * \param etat : le nouvel etat
 * \param mot : l'entree correspondante au processus ds la table des processus
 * \return l'ancien etat 
 */
int changer_etat(int etat, mot_t mot) {
  int old_etat;

  old_etat = mot[2] >> 1;
  mot[2] = (mot[2] & 0xF1) | (etat * 2);
  return old_etat;
}

/**
 * Fonction qui permet de lire l'etat d'un processus
 * L'information est stockee sur les bits 2, 3 et 4 par ordre de poids croissant
 * (00001110) du 2eme octet de poids faible (ie le 3eme en partant de la droite) 
 * du mot
 * \param mot : l'entree correspondante au processus ds la table des processus
 * \return l'etat courant
 */
int lire_etat(mot_t mot) {
  return ((mot[2] & 0x0E) >> 1);
}

/**
 * Fonction qui permet de changer le type de la page physique ou est stockee la 
 * table des pages d'un processus.
 * L'information est stockee ds le bit de poids faible du 2eme octet de poids 
 * faible (ie le 3eme en partant de la droite) du mot
 * \param type : le nouveau type
 * \param mot : l'entree correspondante au processus ds la table des processus
 * \return l'ancien type
 */
int changer_type_pp(int type, mot_t mot) {
  int old_type;

  old_type = (mot[2] & 0x01) + 1;
  mot[2] = (mot[2] & 0xFE) | ((type+1) % 2);
  return old_type;
}

/**
 * Fonction qui permet de lire le type de la page physique ou est stockee la 
 * table des pages d'un processus.
 * L'information est stockee ds le bit de poids faible du 2eme octet de poids 
 * faible (ie le 3eme en partant de la droite) du mot
 * \param mot : l'entree correspondante au processus ds la table des processus
 * \return Le type de la page (PP_SWAP ou PP_RAM)
 */
int lire_type_pp(mot_t mot) {
  return (mot[2] & 0x01) + 1;
}

/**
 * Fonction retournant un numero de processus libre
 * \return 
 * - -1 en cas d'erreur
 * - -2 S'il n'y a plus de pid libre
 * - un num de processus libre
 */
int pid_libre ( ) {
  mot_t mot,mot2;
  int pid_libre;
  int i;

  // Lit le mot ou est stocke le plus petit pid libre
  if (lire_mot_mem(0, NUM_PAGE_META, mot) == -1)
    return -1;
  // Verifie qu'il reste au moins un pid libre
  if ((pid_libre=(int)mot[3]) == NO_MORE_PID) {
    fprintf(stderr,"Il n'y a plus de pid libre\n");
    return -2;
  }
  // Ajoute ce pid a la liste des processus eligibles
  liste_add_pid(liste_proc,pid_libre);
  // Recherche le prochain pid libre
  for (i=pid_libre+1 ; i<N_MAX_PROC+1 ; i++) {
    if (lire_mot_mem(i, NUM_PAGE_META, mot2) == -1) {
      liste_supp_pid(liste_proc,pid_libre);
      return -1;
    }
    if (lire_etat(mot2) == E_MORT) {
      // Met a jour le mot ou est stocke le plus petit pid libre 
      mot[3] = (octet_t)i;
      if (ecrire_mot_mem(0, NUM_PAGE_META, mot) == -1) {
        liste_supp_pid(liste_proc,pid_libre);
        return -1;
      }
      // Renvoie un pid libre
      return pid_libre;
    }
  }
  // il n'y a pas d'autre pid libre
  // Met a jour le mot ou est stocke le plus petit pid libre 
  mot[3] = (octet_t)NO_MORE_PID;
  if (ecrire_mot_mem(0, NUM_PAGE_META, mot) == -1) {
    liste_supp_pid(liste_proc,pid_libre);
    return -1;
  }
  // Renvoie un pid libre
  return pid_libre;
}

/**
 * Fonction permettant de liberer un num de processus apres sa mort
 * \param pid : Numero du processus mort qu'il faut donc declare comme libre
 * \return 
 * - -1 erreur : le numero de processus n'a pas ete libere
 * - -2 erreur : le numero de processus a ete libere mais le numero de le plus petit pid
 *               n'est peut-etre plus exact
 * - 0  succes
 */
int liberer_pid (int pid) {
  mot_t mot,mot2;

  // Libere le num de pid
  changer_etat(E_MORT,mot);
  if (ecrire_mot_mem(pid,NUM_PAGE_META, mot2) == -1)
    return -1;
  // Modifie eventuellement le num du plus petit pid libre
  if (lire_mot_mem(0, NUM_PAGE_META, mot) == -1)
    return -2;
  if (pid < (int)mot[3]) {
    mot[3] = (octet_t)pid;
    if (ecrire_mot_mem(0,NUM_PAGE_META, mot) == -1)
      return -2;
  }
  return 0;
}

/**
 * Fonction retournant un num de page physique libre ds la memoire afin
 * de l'attacher a un processus.
 * \param pid : Numero du processus auquel on veut attacher cette page
 * \return 
 * - -1 en cas d'erreur
 * - -2 S'il n'y a plus de page libre en memoire principale
 * - un num de page libre
 */
int page_libre (int pid) {
  int page_libre;
  int i,j;
  int num_mot, num_octet;
  mot_t mot, mot2;

  // Lit le mot ou est stocke le plus petit pid libre
  if (lire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
    return -1;
  // Verifie qu'il reste au moins une page libre
  if ((page_libre=(int)mot[0]) == NO_MORE_PAGE) {
    fprintf(stdout,"Rem : Il n'y a plus de page libre en memoire RAM\n");
    return -2;
  }
  // Recherche la prochaine page libre
  num_mot = (page_libre + 1) / T_MOT + DEBUT_T_PAGES;
  num_octet = (page_libre + 1) % T_MOT;
  for (i=num_mot ; i<NMPPP ; i++) {
    if (lire_mot_mem(i, NUM_PAGE_META, mot2) == -1)
      return -1;
    for (j=num_octet ; j<T_MOT ; j++) {
      if (mot2[j] == PAGE_VIDE) {
        // Met a jour le mot ou est stocke le plus petit pid libre 
        mot[0] = (octet_t)(i - DEBUT_T_PAGES) * T_MOT + j;
        if (ecrire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
          return -1;
        // Ajoute la page retournee a la liste des pages memoire utilisees
        liste_add_page(liste_pages,page_libre);
        // Attache la page au processus passe en parametre
        num_mot = page_libre / T_MOT + DEBUT_T_PAGES;
        num_octet = page_libre % T_MOT;
        if (lire_mot_mem(num_mot, NUM_PAGE_META, mot2) == -1) {
          liberer_page(page_libre);
          return -1;
        }
        mot2[num_octet] = (octet_t)pid;
        if (ecrire_mot_mem(num_mot, NUM_PAGE_META, mot2) == -1) {
          liberer_page(page_libre);
          return -1;
        }
        return page_libre;
      }
    }
    num_octet = 0;
  }
  // il n'y a pas d'autre page libre
  // Met a jour le mot ou est stocke le plus petit pid libre 
  mot[0] = (octet_t)NO_MORE_PAGE;
  if (ecrire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
    return -1;
  // Ajoute la page retournee a la liste des pages memoire utilisees
  liste_add_page(liste_pages,page_libre);
  // Attache la page au processus passe en parametre
  num_mot = page_libre / T_MOT + DEBUT_T_PAGES;
  num_octet = page_libre % T_MOT;
  if (lire_mot_mem(num_mot, NUM_PAGE_META, mot2) == -1) {
    liberer_page(page_libre);
    return -1;
  }
  mot2[num_octet] = (octet_t)pid;
  if (ecrire_mot_mem(num_mot, NUM_PAGE_META, mot2) == -1) {
    liberer_page(page_libre);
    return -1;
  }
  return page_libre;
}

/**
 * Fonction retournant un num de page physique de la memoire principale.
 * La page retournee est 
 * - soit une page libre s'il reste des pages en memoire principale
 * - soit une page utilisee dont on a copier le contenu ds le swap
 *   apres modification de la table des pages du processus auquel elle appartenait.
 *   Dans ce cas, il s'agit de la page utilisee le moins recemment.
 * \param pid : Numero du processus auquel on veut attacher cette page
 * \return 
 * - -1 en cas d'erreur
 * - -2 S'il n'y a plus de page libre en memoire principale et secondaire
 * - un num de page libre
 */
int page_forcee (int pid) {
  int num_page_mem,num_page_swap,num_table_pages;
  int num_mot_modif,num_mot;
  int old_pid;
  int find_page=0;
  mot_t mot_tab_page,mot_proc;
  int i,j;

  // Regarde s'il y a des pages libres en mem
  num_page_mem = page_libre(pid);
  if (num_page_mem > 0) { // Il reste au moins une page libre en mem 
    return num_page_mem;
  }
  else if (num_page_mem != -2) { // Erreur inconnue
    return -1;
  }
  else { // Il n'y a plus de page libre en mem
    // On cherche la page la moins recemment utilisee
    if ((num_page_mem=liste_old_page(liste_pages)) == -1)
      return -1;
    // On met cette page en swap
    num_page_swap = page_swap_libre();
    if (num_page_swap == -1) // Erreur inconnue
      return -1;
    else if (num_page_swap == -2) // Il n'y a plus de page swap libre
      return -2;
    if (page_mem2swap(num_page_mem,num_page_swap) == -1)
      return -1;
    // On cherche le num du processus auquel appartient la page
    if (lire_mot_mem(DEBUT_T_PAGES+num_page_mem/T_MOT,NUM_PAGE_META,mot_tab_page) == -1) {
      // Libere la page swap allouee
      liberer_page_swap(num_page_swap);
      return -1;
    }
    old_pid =(int) mot_tab_page[num_page_mem%T_MOT];
    // On modifie la table des pages du processus associe a cette page
    if (lire_mot_mem(old_pid,NUM_PAGE_META,mot_proc) == -1) {
      // Libere la page swap allouee
      liberer_page_swap(num_page_swap);
      return -1;
    }
    num_table_pages = (int)mot_proc[3];
    if (num_table_pages == num_page_mem) {
      // La page mise en swap correspond a la table des pages du processus
      mot_proc[3] = (octet_t)num_page_swap;
      changer_type_pp(PP_SWAP,mot_proc);
      if (ecrire_mot_mem(old_pid,NUM_PAGE_META,mot_proc) == -1) {
        // Libere la page swap allouee
        liberer_page_swap(num_page_swap);
        return -1;
      }
      // Change le num du processus proprietaire de la page
      mot_tab_page[num_page_mem%T_MOT] = (octet_t)pid;
      if (ecrire_mot_mem(DEBUT_T_PAGES+num_page_mem/T_MOT,NUM_PAGE_META,mot_tab_page) == -1) {
        // Reviens a l'ancien etat de la table des processus
        mot_proc[3] = (octet_t)num_page_mem;
        changer_type_pp(PP_RAM,mot_proc);
        ecrire_mot_mem(old_pid,NUM_PAGE_META,mot_proc);
        // Libere la page swap allouee
        liberer_page_swap(num_page_swap);
        return -1;
      }
    }
    else {
      // La page est une page de code, de donnees, de pile ou de tampon
      for (i=0 ; (i<3) && (find_page==0) ; i++) {
        for (j=0 ; j<T_MAX_ZONES ; j++) {
          if (lire_mot_mem(i*8+j,num_table_pages,mot_proc) == -1) {
            // Libere la page swap allouee
            liberer_page_swap(num_page_swap);
            return -1;
          }
          if ((int)mot_proc[2] == PP_NULL)
            // On passe a la zone suivante
            break;
          if (mot_proc[3] == num_page_mem) {
            // On a trouve ds la table des pages du processus, l'entree correspondant a 
            // la page mise en swap. Il faut la modifier en consequence
            mot_proc[2] = (octet_t)PP_SWAP;
            mot_proc[3] = (octet_t)num_page_swap;
            if (ecrire_mot_mem(i*8+j,num_table_pages,mot_proc) == -1) {
              // Libere la page swap allouee
              liberer_page_swap(num_page_swap);
              return -1;
            }
            find_page = 1;
            num_mot_modif = i*8+j;
            break;
          }
        }
      }
      if (find_page == 0) {
        // La page est peut etre une page de tampon
        num_mot = NUM_MOT_IN;
        if (lire_mot_mem(num_mot,num_table_pages,mot_proc) == -1) {
          liberer_page_swap(num_page_swap);
          return -1;
        }
        if (((int)mot_proc[3] != num_page_mem) || ((int)mot_proc[2] != PP_SWAP)) {
          num_mot = NUM_MOT_OUT;
          if (lire_mot_mem(num_mot,num_table_pages,mot_proc) == -1) {
            liberer_page_swap(num_page_swap);
            return -1;
          }
          if (((int)mot_proc[3] != num_page_mem) || ((int)mot_proc[2] != PP_SWAP)) {
            // La page rapatriee n'appartient pas au processus 'pid'
            fprintf(stderr,"La table des pages du processus %d n'est pas coherente avec la ",old_pid);
            fprintf(stderr,"table des pages de la memoire principale.\n");
            liberer_page_swap(num_page_swap);
            return -1;
          }
        }
        mot_proc[2] = (octet_t)PP_SWAP;
        mot_proc[3] = (octet_t)num_page_swap;
        num_mot_modif = num_mot;
        if (ecrire_mot_mem(num_mot,num_table_pages,mot_proc) == -1) {
          liberer_page_swap(num_page_swap);
          return -1;
        }
      }
      // Change le num du processus proprietaire de la page
      mot_tab_page[num_page_mem%T_MOT] = (octet_t)pid;
      if (ecrire_mot_mem(DEBUT_T_PAGES+num_page_mem/T_MOT,NUM_PAGE_META,mot_tab_page) == -1) {
        // Reviens a l'ancien etat de la table des pages du processus
        mot_proc[3] = (octet_t)num_page_mem;
        mot_proc[2] = (octet_t)PP_RAM;
        ecrire_mot_mem(num_mot_modif,num_table_pages,mot_proc);
        // Libere la page swap allouee
        liberer_page_swap(num_page_swap);
        return -1;
      }
    }
  }
  liste_acces_page(liste_pages,num_page_mem);
  return num_page_mem;
}

/**
 * Fonction permettant de liberer un numero de page physique ds la memoire
 * \param page : Numero de la page non utilisee (a liberer)
 * \return 
 * - -1 erreur : le numero de page n'a pas ete libere
 * - -2 erreur : le numero de page a ete libere mais le numero de la plus petite page
 *               n'est peut-etre plus exact
 * - 0  succes
 */
int liberer_page(int page) {
  mot_t mot;
  int num_mot,num_octet;

  // Libere le num de la page
  num_mot = page/T_MOT + DEBUT_T_PAGES;
  num_octet = page % T_MOT;
  if (lire_mot_mem(num_mot, NUM_PAGE_META, mot) == -1)
    return -1;
  mot[num_octet] = PAGE_VIDE;
  if (ecrire_mot_mem(num_mot,NUM_PAGE_META, mot) == -1)
    return -1;
  // Modifie eventuellement le num de la plus petite page libre
  if (lire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
  return -2;
  if (page < (int)mot[0]) {
    mot[0] = (octet_t)page;
    if (ecrire_mot_mem(DEBUT_T_PAGES,NUM_PAGE_META, mot) == -1)
      return -2;
  }
  // Supprime cette page de la liste des pages memoire
  // Mise a jour de la liste des pages memoire
  if (liste_supp_page(liste_pages,page) != 0) {
    fprintf(stderr,"La table des pages memoire utilisees n'etait pas coherente.\n");
    // On continue tout de meme car elle est peut-etre de nouveau coherente
  }
  return 0;
}

/**
 * Fonction retournant un numero de page physique du swap libre 
 * \return 
 * - -1  En cas d'erreur
 * - -2 S'il n'y a plus de page libre en memoire principale
 * - un num de page libre
 */
int page_swap_libre () {
  mot_t mot,mot2;
  int pc,num_mot,num_page_libre;

  // Lecture de la position courante ds la table des pages swap libres
  if (lire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
    return -1;
  pc = (int)mot[1];
  // Lecture du num de la page swap libre
  num_mot = pc / T_MOT;
  if (lire_mot_mem(num_mot, NUM_PAGE_SWAP, mot2) == -1) 
    return -1;
  num_page_libre = (int)mot2[pc%T_MOT];
  if (num_page_libre == NO_MORE_SWAP) {
    fprintf(stderr,"Il n'y a plus de page libre dans le swap\n");
    return -2;
  }
  // Deplacement du pointeur courant
  pc ++;
  // Rechargement du debut de la liste des pages swap libres si necessaire
  if (pc >= NMPPP*T_MOT) {
    if (page_swap2mem(num_page_libre, NUM_PAGE_SWAP) == -1)
      return -1;
    mot[1] = (octet_t)0;
    if (ecrire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1) {
      fprintf(stderr,"Position courante de la table des pages swap libres incoherente\n");
      return -1;
    }
  }
  else {
    // Modification de la position courante
    mot[1] = (octet_t)pc;
    if (ecrire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
      return -1;
  }
  return num_page_libre;
}

/**
 * Fonction permettant de liberer un numero de page physique ds le swap
 * \param page : Numero de la page non utilisee (a liberer)
 * \return 
 * - -1 erreur : le numero de page n'a pas ete libere
 * - -2 erreur : le numero de page a ete libere mais le pointeur courant n'a pas ete mis a jour
 * - 0  succes
 */
int liberer_page_swap (int page) {
  mot_t mot,mot2;
  int pc,num_mot;

  // Lecture de la position courante ds la table des pages swap libres
  if (lire_mot_mem(DEBUT_T_PAGES, NUM_PAGE_META, mot) == -1)
    return -1;
  pc = (int)mot[1];
  if (pc > 0) {
    pc --;
    num_mot = pc / T_MOT;
    if (lire_mot_mem(num_mot, NUM_PAGE_SWAP, mot2) == -1) 
      return -1;
    mot2[pc%T_MOT] = (octet_t)page;
    // Libere le num de la page
    if (ecrire_mot_mem(num_mot,NUM_PAGE_SWAP, mot2) == -1)
      return -1;
    // Modifie le pointeur courant
    mot[1] = (octet_t)pc;
    if (ecrire_mot_mem(DEBUT_T_PAGES,NUM_PAGE_META, mot) == -1)
      return -2;
  }
  else {
    // Sauvegarde dans le swap de cette page de la liste des pages swap libres
    if (page_mem2swap(NUM_PAGE_SWAP,page) == -1)
      return -1;
    // Creation de la nouvelle page de la liste des pages swap libres
    mot2[3] = (octet_t)page;
    if (ecrire_mot_mem(NMPPP-1,NUM_PAGE_SWAP,mot2) == -1)
      return -1;
    pc = NMPPP*T_MOT-1;
    mot[1] = (octet_t)pc;
    if (ecrire_mot_mem(DEBUT_T_PAGES,NUM_PAGE_META, mot) == -1)
      return -2;
  }
  return 0;
}

/**
 * Fonction permettant de mettre en memoire principale une page du processus 'pid'
 * se trouvant dans le swap.
 * Fait toutes les modifications necessaires dans la table des pages RAM et la table
 * des processus ou la table des pages du processus 'pid'.
 * \param page : Numero de la page swap qu'on veut mettre en memoire principale
 * \return 
 * - -1 en cas d'erreur
 * - -2 erreur : plus assez de memoire pour realiser l'operation (minimum : 1  page swap libre)
 * - le numero de la page memoire dans laquelle a ete copiee la page swap
 */
int rapatrier_page_swap (int pid, int page) {
  int num_page_mem,num_tab_pages;
  int num_mot;
  int i,j;
  int find_page = 0;
  mot_t mot_proc,mot_tab;

  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
    return -1;
  }
  if (lire_etat(mot_proc) == E_MORT) {
    fprintf(stderr,"Impossible de rapatrier une page d'un processus mort (%d)\n",pid);
    return -1;
  }
  // Regarde s'il y a des pages libres en mem
  num_page_mem = page_forcee(pid);
  if (num_page_mem > 0) {
    // Rapatrie la page swap ds cette page libre
    if (page_swap2mem(page, num_page_mem) == -1) {
      liberer_page(num_page_mem);
      return -1;
    }
  }
  else if (num_page_mem != -2) {
    return -1;
  }
  else {
    // Il n'y a plus de page physique libre
    fprintf(stderr,"Memoire insuffisante\n");
    return -2;
  }
  // Modifie le mot qui pointait sur 'page'
  // Regarde a quel type de page correspond la page rapatriee
  if ((mot_proc[3] == page) && (lire_type_pp(mot_proc) == PP_SWAP)) {
    // La page rapatriee correspond a la table des pages du proc
    mot_proc[3] = (octet_t)num_page_mem;
    changer_type_pp(PP_RAM,mot_proc);
    if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
      liberer_page(num_page_mem);
      return -1;
    }
  }
  else { 
    // La page rapatriee est une page de code, pile, donnees ou tampon
    switch (lire_type_pp(mot_proc)) {
    case PP_SWAP:
      switch (num_tab_pages=rapatrier_page_swap(pid,(int)mot_proc[3])) {
      case -1:
        liberer_page(num_page_mem);
        return -1;
      case -2:
        liberer_page(num_page_mem);
        return -2;
      }
      break;
    case PP_RAM:
      num_tab_pages = (int)mot_proc[3];
      break;
    }
    // Parcours la table des pages a la recherche de la page rapatriee
    for (i=0 ; (i<3) && (find_page==0) ; i++) {
      for (j=i*T_MAX_ZONES ; (j<(i+1)*T_MAX_ZONES) && (find_page==0) ; j++) {
        if (lire_mot_mem(j,num_tab_pages,mot_tab) == -1) {
          liberer_page(num_page_mem);
          return -1;
        }
        if (((int)mot_tab[3] == page) && ((int)mot_tab[2] == PP_SWAP)) {
          // On a trouve le mot qui pointe sur la page rapatriee
          find_page = 1;
          mot_tab[3] = (octet_t)num_page_mem;
          mot_tab[2] = (octet_t)PP_RAM;
          if (ecrire_mot_mem(j,num_tab_pages,mot_tab) == -1) {
            liberer_page(num_page_mem);
            return -1;
          }
        }
        else if ((int)mot_tab[2] == PP_NULL)
          break;
      }
    }
    if (find_page == 0) {
      // La page rapatriee est une page de tampon
      num_mot = NUM_MOT_IN;
      if (lire_mot_mem(num_mot,num_tab_pages,mot_tab) == -1) {
        liberer_page(num_page_mem);
        return -1;
      }
      if (((int)mot_tab[3] != page) || ((int)mot_tab[2] != PP_SWAP)) {
        num_mot = NUM_MOT_OUT;
        if (lire_mot_mem(num_mot,num_tab_pages,mot_tab) == -1) {
          liberer_page(num_page_mem);
          return -1;
        }
        if (((int)mot_tab[3] != page) || ((int)mot_tab[2] != PP_SWAP)) {
          // La page rapatriee n'appartient pas au processus 'pid'
          fprintf(stderr,"La page %d en swap n'appartient pas au proc %d\n",page,pid);
          liberer_page(num_page_mem);
          return -1;
        }
      }
      mot_tab[3] = (octet_t)num_page_mem;
      mot_tab[2] = (octet_t)PP_RAM;
      if (ecrire_mot_mem(num_mot,num_tab_pages,mot_tab) == -1) {
        liberer_page(num_page_mem);
        return -1;
      }
    }
  }
  // Libere la page swap ramenee en memoire principale
  if (liberer_page_swap(page) == -1) 
    return -1;
  return num_page_mem;
}

/**
 * Fonction permettant de lire un mot se trouvant ds la mem principale
 * \param num_pp : Numero de la page physique en memoire ds laquelle on veut lire un mot
 * \param num_mot : Numero du mot qu'on veut lire ds cette page
 * \param mot : Structure qui sera remplie apres lecture
 * \return 
 * - -1 en cas d'erreur
 * -  0 sinon
 */
int lire_mot_mem(int num_mot, int num_pp, mot_t mot) {
  unsigned char *p_mot;

  // Verifie que les coordonnees du mot sont valides
  if (num_pp < 0 || num_pp > NPP-1) {
    fprintf(stderr,"%d n'est pas un numero de page physique en memoire correct\n",num_pp);
    return -1;
  }
  if (num_mot < 0 || num_mot > NMPPP-1) {
    fprintf(stderr,"%d n'est pas un numero de mot correct\n",num_mot);
    return -1;
  }
  // Lecture du mot
  p_mot = mem[num_pp][num_mot];
  // Remplissage de la structure mot_t
  memcpy(mot,p_mot,sizeof(mot_t));
  // Mise a jour de la liste des pages memoire
  if ((num_pp != NUM_PAGE_META) && (num_pp != NUM_PAGE_SWAP)) {
    if (liste_acces_page(liste_pages,num_pp) != 0) {
      fprintf(stderr,"Tentative de lecture ds une page non reservee (%d)\n",num_pp);
      return -1;
    }
  }
  return 0;
}

/**
 * Fonction permettant d'ecrire un mot dans la mem principale
 * \param num_pp : Numero de la page physique en memoire ds laquelle on veut ecrire un mot
 * \param num_mot : Emplacement ds cette page ou on veut ecrire le mot
 * \param mot : Le mot a ecrire
 * \return 
 * - -1 en cas d'erreur
 * -  0 sinon
 */
int ecrire_mot_mem(int num_mot, int num_pp, mot_t mot) {
  unsigned char *p_mot;

  // Verifie que les coordonnees du mot sont valides
  if (num_pp < 0 || num_pp > NPP-1) {
    fprintf(stderr,"%d n'est pas un numero de page physique en memoire correct\n",num_pp);
    return -1;
  }
  if (num_mot < 0 || num_mot > NMPPP-1) {
    fprintf(stderr,"%d n'est pas un numero de mot correct\n",num_mot);
    return -1;
  }
  // Ecriture du mot
  p_mot = mem[num_pp][num_mot];
  memcpy(p_mot,mot,sizeof(mot_t));
  // Mise a jour de la liste des pages memoire
  if ((num_pp != NUM_PAGE_META) && (num_pp != NUM_PAGE_SWAP)) {
    if (liste_acces_page(liste_pages,num_pp) != 0) {
      fprintf(stderr,"Tentative d'ecriture ds une page non reservee (%d)\n",num_pp);
      return -1;
    }
  }
  return 0;
}

/**
 *  Fonction retournant a partir d'une adresse virtuelle :
 * - le numero de la page physique a laquelle elle correspond
 * - le type (memoire principale ou secondaire) de cette page
 * \param pid : num du processus auquel appartient l'adresse virtuelle
 * \param ad_v : adresse virtuelle dont on veut connaitre l'adresse physique correspondante
 * \param info : structure qui sera remplie par l'adresse physique, son type et ses droits
 * \return 
 * - -1 en cas d'erreur
 * - -2 recherche en dehors de la pile
 * - 0  sinon
 */
int adresse_phys (int pid, vad_t ad_v, info_mot_t *info) {
  int num_page, num_mot;
  mot_t mot;
  int taille;

  // Verifie la coherence de l'adresses virtuelle
  if ((ad_v.zone!=Z_CODE) && (ad_v.zone!=Z_PILE) && (ad_v.zone!=Z_DONNEES)) {
    fprintf(stderr,"Adresse virtuelle incoherente: zone inconnue\n");
    return -1;
  }
  if ((ad_v.num < 0) || (ad_v.num > T_MAX_ZONES*NMPPP)) {
    fprintf(stderr,"Adresse virtuelle incoherente: depassement de zone\n");
    return -1;
  }
  // Lecture de la table des pages du processus pid
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Tentative de lecture d'une page d'un proc mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_page=rapatrier_page_swap(pid,(int)mot[3])) == -1)
      return -1;
    else if (num_page == -2)
      // Il n'y a plus assez de mem pour lire la table des pages du processus
      // Impossible de continuer
      return -1;
    break;
  case PP_RAM:
    num_page = (int)mot[3];
    break;
  }
  // Regarde dans quelle zone se trouve l'adresse car le traitement des adresses de la pile 
  // ne se fait pas de la meme facon que pour celles de code ou de donnees
  if (ad_v.zone != Z_PILE) {
    num_mot = ad_v.zone*T_MAX_ZONES+ad_v.num/NMPPP;
    if (lire_mot_mem(num_mot, num_page, mot) == -1)
      return -1;
    // Remplissage de la structure info_mot_t
    info->type = (int)mot[2];
    info->page = (int)mot[3];
    info->mot = ad_v.num % NMPPP;
    return 0;
  }
  else {
    // Dans la pile les adresses se lisent a l'envers (P0 = haut de la pile)
    // Cherche le haut de la pile
    if ((taille=lire_esp(pid)) == -1)
      return -1;
    if (taille == 0) {
      // L'adresse virtuelle est incorrecte: il n'y a aucune donnees ds la pile
      fprintf(stderr,"Recherche de donnees ds une pile vide!\n");
      return -2;
    }
    if ((taille-ad_v.num) < 1) {
      // L'adresse virtuelle est incorrecte: reference un donnee hors de la pile
      fprintf(stderr,"Recherche de donnees en dehors de la pile\n");
      return -2;
    }
    num_mot = ad_v.zone*T_MAX_ZONES + ((taille-1) - ad_v.num)/NMPPP;
    if (lire_mot_mem(num_mot, num_page, mot) == -1)
      return -1;
    // Remplissage de la structure info_mot_t
    info->type = (int)mot[2];
    info->page = (int)mot[3];
    info->mot = ((taille-1) - ad_v.num) % NMPPP;
    return 0;
  }
}

/**
 * Fonction permettant d'ecrire une page dans la memoire swap
 * \param page : page a ecrire
 * \param num_page : numero de la page swap ou il faut ecrire
 * \return 
 * - -1 en cas d'erreur
 * - 0  sinon
 */
int ecrire_page_swap (page_t page, int num_page) {
  if (lseek(swap_d,num_page*sizeof(page_t),SEEK_SET) == -1)
    return -1;
  if (write(swap_d,page,sizeof(page_t)) != sizeof(page_t))
    return -1;
  return 0;
}

/**
 * Fonction permettant de copier une page du swap dans la memoire princale
 * \param num_page_swap : numero de la page a copier
 * \param num_page_mem : numero de la page memoire dans laquelle on veut faire la copie
 * \return 
 * - -1 en cas d'erreur
 * - 0  sinon
 */
int page_swap2mem (int num_page_swap, int num_page_mem) {
  if (lseek(swap_d,num_page_swap*T_PAGE,SEEK_SET) == (off_t)-1)
    return -1;
  if (read(swap_d,mem[num_page_mem],T_PAGE) != T_PAGE)
    return -1;
  return 0;
}

/**
 * Fonction permettant de copier une page de la mem princale ds le swap
 * \param num_page_mem : numero de la page a copier
 * \param num_page_swap : numero de la page swap dans laquelle on veut faire la copie
 * \return 
 * - -1 en cas d'erreur
 * - 0  sinon
 */
int page_mem2swap (int num_page_mem, int num_page_swap) {
  if (lseek(swap_d,num_page_swap*T_PAGE,SEEK_SET) == -1)
    return -1;
  if (write(swap_d,mem[num_page_mem],T_PAGE) != T_PAGE)
    return -1;
  return 0;
}

/**
 * Fonction permettant d'ajouter une page a la table des pages d'un processus.
 * Cette page est ajoutee a la suite des pages deja presentes.
 * \param pid : numero du processus auquel on veut ajouter une page
 * \param zone : zone memoire du processus ou la page doit etre ajoutee
 * \param page : numero de la page en memoire principale a ajouter
 * \return 
 * - -1 en cas d'erreur
 * - -2 plus de place pour rajouter une page ds la zone voulue
 * - 0  en cas de succes
 */
int page_add2proc (int pid, int zone, int page) {
  int num_page, num_mot;
  mot_t mot;
  int i;

  // Cherche le num de la table des pages du processus pid
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Tentative de lecture d'une page d'un proc mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_page=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  case PP_RAM:
    num_page = (int)mot[3];
    break;
  }
  // Lecture de la table des pages
  num_mot = zone*T_MAX_ZONES;
  for (i=0 ; i<T_MAX_ZONES ; i++) {
    if (lire_mot_mem(num_mot, num_page, mot) == -1)
      return -1;
    if ((int)mot[2] == PP_NULL) {
      // 1er entree non attribuee ds cette table des pages du processus
      // On y met la page a ajouter
      mot[2] = (octet_t) PP_RAM;
      mot[3] = (octet_t) page;
      if (ecrire_mot_mem(num_mot, num_page, mot) == -1)
        return -1;
      return 0;
    }
    num_mot++;
  }
  // Il n'y a plus de place ds cette zone de la memoire du processus
  fprintf(stderr,"Il n'y a plus de place ds zone %d de memoire du processus %d\n",zone,pid);
  return -2;
}

/**
 * Fonction permettant de supprimer la derniere page d'une des zones memoire 
 * d'un processus. Cela consiste a:
 * - supprimer le mot qui lui correspond dans la table des pages du processus
 * - liberer cette page
 * \param pid : numero du processus auquel on veut supprimer une page
 * \param zone : zone memoire du processus ou la derniere page doit etre suprimer
 * \return 
 * - -1 en cas d'erreur
 * - -2 il n'y a pas de page a supprimer dans cette zone
 * - 0  en cas de succes
 */
int page_supp2proc (int pid, int zone) {
  int num_page, num_mot;
  mot_t mot,mot_precedent;

  // Lecture de la table des pages du processus pid
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    // Ce num de processus ne correspond a aucune page physique
    fprintf(stderr,"Le processus %d n'est pas utilise\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_SWAP:
    // La table des pages du processus pid se trouve ds le swap
    // On la ramene en mem principale
    if ((num_page=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  case PP_RAM:
    num_page = (int)mot[3];
    break;
  }
  // Suppression de la derniere page
  for (num_mot=zone*T_MAX_ZONES ; num_mot<(zone+1)*T_MAX_ZONES ; num_mot++) {
    if (lire_mot_mem(num_mot, num_page, mot) == -1)
      return -1;
    if ((int)mot[2] == PP_NULL) {
      // 1er entree non attribuee ds cette table des pages du processus
      if (num_mot%T_MAX_ZONES == 0) {
        // La zone dans laquelle on veut retirer une page n'en contient aucune
        fprintf(stderr,"Il n'y a pas de page a supprimer ds zone %d de memoire du processus %d\n",zone,pid);
        return -2;
      }
      // Si cette entree est vide, alors il faut supprimer l'entree precedente
      mot[2] = (octet_t) PP_NULL;
      if (ecrire_mot_mem(num_mot-1, num_page, mot) == -1)
        return -1;
      // Sort du for pour liberer la page
      break;
    }
    else {
      // Ce mot correspond a une page utilisee. Il s'agit peut etre de la derniere page
      // de la zone. Alors on memorise ses coordonnees.
      memcpy(mot_precedent,mot,sizeof(mot_t));
    }
  }
  if (num_mot == (zone+1)*T_MAX_ZONES) {
    // Le mot a supprimer est le dernier
    mot[2] = (octet_t) PP_NULL;
    if (ecrire_mot_mem(num_mot-1, num_page, mot) == -1)
      return -1;
  }
  // Liberation de la page
  switch ((int)mot_precedent[2]) {
  case PP_RAM:
    if (liberer_page((int)mot_precedent[3]) != 0) {
      // Remet la table des pages ds son etat premier
      ecrire_mot_mem(num_mot-1, num_page, mot_precedent);
      return -1;
    }
    return 0;
  case PP_SWAP:
    if (liberer_page_swap((int)mot_precedent[3]) != 0) {
      // Remet la table des pages ds son etat premier
      ecrire_mot_mem(num_mot-1, num_page, mot_precedent);
      return -1;
    }
    return 0;
  default:
    fprintf(stderr,"Table des pages du processus %d incoherente\n",pid);
    return -1;
  }
}

/**
 * Fonction permettant de lire un mot a partir d'une adresse virtuelle (si elle correspond a
 * une page physique existante).
 * \param pid : numero du processus auquel appartient la page dont on veut lire un mot
 * \param ad_v : adresse virtuelle du mot a lire
 * \param mot : structure dans laquelle on stocke le mot lu
 * \return 
 * - -1 en cas d'erreur
 * - -2 le mot ne fait pas parti de l'espace mem du processus (cependant il a qd meme ete lu
 *      si il fait parti de la zone de code ou de donnees)
 * - 0  sinon
 */
int lire_mot (int pid, vad_t ad_v, mot_t mot) {
  info_mot_t info;
  int num_page;

  // Recherche l'adresse physique du mot
  switch (adresse_phys (pid, ad_v, &info)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Rapatrie eventuellement le mot en memoire principale
  switch (info.type) {
  case PP_SWAP:
    if ((num_page=rapatrier_page_swap(pid,info.page)) < 0)
      return -1;
    break;
  case PP_RAM:
    num_page = info.page;
    break;
  case PP_NULL:
    fprintf(stderr,"Tentative de lecture d'une page inexistante\n");
    return -1;
  default:
    fprintf(stderr,"Table des pages du processus %d incoherente\n",pid);
    return -1;
  }
  if (lire_mot_mem(info.mot,num_page,mot) == -1)
    return -1;
  // Verifie que le mot fait bien parti de l'espace mem du processus
  if ((int)mot[0] == M_NU)
    return -2;
  return 0;
}

/**
 * Fonction qui ecrit un mot a partir d'une adresse virtuelle. Il faut que cette adresse virtuelle
 * reference une adresse physique existante car elle ne cree pas de page automatiquement.
 * Le mot ecrit est systematiquement marque comme utilise car la fonction ne sert pas a diminuer 
 * l'espace memoire du processus.
 * \param pid : Numero du processus auquel appartient la page dans laquelle on veut ecrire un mot
 * \param ad_v : Adresse virtuelle du mot a ecrire
 * \param mot : Le mot a ecrire
 * \return 
 * - -1 en cas d'erreur
 * - -2 en cas d'ecriture en dehors de l'espace mem
 * - 0  sinon
 */
int ecrire_mot (int pid, vad_t ad_v, mot_t mot) {
  info_mot_t info;
  int num_page;
  mot_t mot_lu;

  // Verifie qu'on a les droits en ecriture
  if (ad_v.zone == Z_CODE) {
    fprintf(stderr,"Impossible de modifier la zone de code\n");
    return -1;
  }
  // Recherche l'adresse physique du mot
  switch (adresse_phys (pid, ad_v,&info)) {
  case -1:
    return -1;
  case -2:
    return -2;
  }
  // Rapatrie eventuellement la page contenant le mot en memoire principale
  switch (info.type) {
  case PP_SWAP:
    if ((num_page=rapatrier_page_swap(pid,info.page)) < 0)
      return -1;
    break;
  case PP_RAM:
    num_page = info.page;
    break;
  case PP_NULL:
    fprintf(stderr,"Tentative d'ecriture dans une page inexistante\n");
    return -1;
  default:
    fprintf(stderr,"Table des pages du processus %d incoherente\n",pid);
    return -1;
  }
  // Verifie que le mot est ds l'espave mem du proc
  if (lire_mot_mem(info.mot,num_page,mot_lu) == -1)
    return -1;
  if ((int)mot_lu[0] == M_NU)
    return -2;
  // Verifie que le mot a ecrire n'est pas marque comme inutilise
  if ((int)mot[0] ==  M_NU)
    mot[0] = M_UT;
  // Ecrit le mot
  if (ecrire_mot_mem(info.mot,num_page,mot) == -1)
    return -1;
  return 0;
}

/**
 * Fonction permettant de modifier le pid du processus pere.
 * Le ppid devrait se trouver dans la table des processus. Cependant, faute de place, il
 * est stockee dans un mot inutilise (le NUM_MOT_PPID+1 eme) de la table des pages de chaque
 * processus.
 * \param pid : num du processus
 * \param ppid : le nouveau num du processus pere
 * \return 
 * - -1 en cas d'erreur
 * - 0  sinon
 */
int changer_ppid (int pid, int ppid) {
  mot_t mot;
  int num_tab_pages;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Acces a un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Ecrit le nouveau num du processus pere
  mot[3] = (octet_t)ppid;
  if (ecrire_mot_mem(NUM_MOT_PPID,num_tab_pages,mot) == -1)
    return -1;
  return 0;
}

/**
 * Fonction permettant de lire le pid du processus pere.
 * Le ppid devrait se trouver dans la table des processus. Cependant, faute de place, il
 * est stockee dans un mot inutilise (le NUM_MOT_PPID+1 eme) de la table des pages de chaque
 * processus.
 * \param pid : num du processus
 * \return 
 * - -1 en cas d'erreur
 * - le ppid  sinon
 */
int lire_ppid (int pid) {
  mot_t mot;
  int num_tab_pages;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Lit le num du processus pere
  if (lire_mot_mem(NUM_MOT_PPID,num_tab_pages,mot) == -1)
    return -1;
  return (int)mot[3];
}

/**
 * Modifie la hauteur de la pile.\n
 * Cette hauteur devrait se trouver dans la table des processus. Cependant, faute de place,
 * elle est stockee dans un mot inutilise (le NUM_MOT_ESP+1 eme) de la table des pages de 
 * chaque processus.\n
 * En plus de modifier la taille, la fonction met aussi a jour la pile en:
 * - marquant le dernier mot de la pile comme non utilise en cas de decrementattion \n
 *   supprime eventuellement une page physique.
 * - marquant le 1er mot au-dela de la pile comme utilise en cas d'incrementation \n
 *   rajoute eventuellement une page physique.
 * \param pid : numero du processus dont on veut connaitre la hauteur de pile
 * \param op : operation (incremente / decremente) a realiser sur la pile
 * \return 
 * - -1 en cas d'erreur
 * - 0 sinon
 */
int changer_esp (int pid, int op) {
  mot_t mot, mot_pile, mot_reg;
  int num_tab_pages;
  int taille,modif;
  int num_p_pile;
  int i;

  // Verifie que l'operation est correcte
  if ((op != 1) && (op != -1)) {
    fprintf(stderr,"Operation sur la pile du processus %d illegale\n",pid);
    return -1;
  }
  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Acces a un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Lit la taille de la pile
  if (lire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg) == -1)
    return -1;
  // Verifie qu'il est possible de realisre l'operation voulue
  taille = (int)mot_reg[3];
  if ((op == 1) && (taille == SP_MAX)) {
    fprintf(stderr,"Impossible d'incrementer la pile du proc %d: pile pleine\n",pid);
    return -1;
  }
  else if ((op == -1) && (taille == 0)) {
    fprintf(stderr,"Impossible de decrementer la pile du proc %d: pile vide\n",pid);
    return -1;
  }
  // Modifie la taille de la pile
  taille += op;
  mot_reg[3] = (octet_t)taille;
  if (ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg) == -1)
    return -1;
  // MAJ de la pile
  if (op == -1) { // Operation de decrementation
    modif = taille;
    // Lit la derniere page de la pile
    if (lire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile) == -1) {
      mot_reg[3] = (octet_t)taille - op;
      ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
      return -1;
    }
    if ((modif % NMPPP) == 0) {
      // On supprime la derniere page de la pile
      switch ((int)mot_pile[2]) {
      case PP_RAM:
        // Modifie la table des pages
        mot_pile[2] = (octet_t)PP_NULL;
        if (ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile) == -1) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          return -1;
        }
        // Libere la page
        if (liberer_page((int)mot_pile[3]) == -1) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          mot_pile[2] = (octet_t)PP_RAM;
          ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile);
          return -1;
        }
        return 0;
      case PP_SWAP:
        // Modifie la table des pages
        mot_pile[2] = (octet_t)PP_NULL;
        if (ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile) == -1) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          return -1;
        }
        // Libere la page
        if (liberer_page_swap((int)mot_pile[3]) == -1) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          mot_pile[2] = (octet_t)PP_SWAP;
          ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile);
          return -1;
        }
        return 0;
      default:
        fprintf(stderr,"Table des pages du proc %d incoherente\n",pid);
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
    }
    else {
      // On marque le dernier mot de la pile comme inutilise
      switch ((int)mot_pile[2]) {
      case PP_RAM:
        num_p_pile = (int)mot_pile[3];
        break;
      case PP_SWAP:
        if ((num_p_pile=rapatrier_page_swap(pid,(int)mot_pile[3])) < 0) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          return -1;
        }
        break;
      default:
        fprintf(stderr,"Table des pages du proc %d incoherente\n",pid);
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
      mot_pile[0]  = (octet_t)M_NU;
      if (ecrire_mot_mem(modif%NMPPP,num_p_pile,mot_pile) == -1) {
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
    }
  }
  if (op == 1) { // Operation d'incrementation
    modif = taille-1;
    if ((modif % NMPPP) == 0) {
      // On rajoute une nouvelle page a la pile
      if ((num_p_pile = page_forcee(pid)) < 0) {
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
      mot_pile[3] = (octet_t)num_p_pile;
      mot_pile[2] = (octet_t)PP_RAM;
      if (ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES, num_tab_pages, mot_pile) == -1) {
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        liberer_page(num_p_pile);
        return -1;
      }
      // Initialise cette nouvelle page de pile
      mot_pile[3] = (octet_t)0;
      mot_pile[2] = (octet_t)0;
      for (i=0 ; i<NMPPP ; i++) {
        if (i == 0)
          mot_pile[0] = (octet_t)M_UT;
        else
          mot_pile[0] = (octet_t)M_NU;
        if (ecrire_mot_mem(i, num_p_pile, mot_pile) == -1) {
          // Annule l'operation sur la taille de la pile
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          // Libere la page allouee
          liberer_page(num_p_pile);
          // Supprime cette page de la table des pages du proc
          mot_pile[2] = (octet_t)PP_NULL;
          ecrire_mot_mem(modif/NMPPP+2*T_MAX_ZONES, num_tab_pages, mot_pile);
          return -1;
        }
      }
      return 0;
    }
    else {
      // Lit la derniere page de la pile pour la modifier
      if (lire_mot_mem(modif/NMPPP+2*T_MAX_ZONES,num_tab_pages,mot_pile) == -1) {
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
      // On marque le dernier mot de la pile agrandie comme utilise
      switch ((int)mot_pile[2]) {
      case PP_RAM:
        num_p_pile = (int)mot_pile[3];
        break;
      case PP_SWAP:
        if ((num_p_pile=rapatrier_page_swap(pid,(int)mot_pile[3])) < 0) {
          mot_reg[3] = (octet_t)taille - op;
          ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
          return -1;
        }
        break;
      default:
        fprintf(stderr,"Table des pages du proc %d incoherente\n",pid);
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
      mot_pile[0]  = (octet_t)M_UT;
      mot_pile[3] = (octet_t)0;
      mot_pile[2] = (octet_t)0;
      if (ecrire_mot_mem(modif%NMPPP,num_p_pile,mot_pile) == -1) {
        mot_reg[3] = (octet_t)taille - op;
        ecrire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot_reg);
        return -1;
      }
    }
  }
  return 0;
}

/**
 * Retourne la hauteur de la pile.
 * Cette hauteur devrait se trouver dans la table des processus. Cependant, faute de place,
 * elle est stockee dans un mot inutilise (le NUM_MOT_ESP+1 eme) de la table des pages de 
 * chaque processus.\n
 * \param pid : numero du processus dont on veut connaitre la hauteur de pile
 * \return 
 * - -1 en cas d'erreur
 * - la hauteur sinon
 */
int lire_esp (int pid) {
  mot_t mot;
  int num_tab_pages;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Acces a un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Lit la taille de la pile
  if (lire_mot_mem(NUM_MOT_ESP,num_tab_pages,mot) == -1)
    return -1;
  return (int)mot[3];
}

/**
 * Verrouille un tampon:
 * - Verifie que le tampmon n'est pas deja verrouille
 * - Vourrouille le tampon
 * \param pid : num du processus
 * \param num_p_tampon : num du mot ds la table des pages du processus correspondant au tampon
 * \return 
 * - -1 en cas d'erreur
 * - -3 le tampon est deja verrouille
 * - 0  sinon
 */
int verrouiller_tampon (int pid, int num_tampon) {
  mot_t mot;
  int num_tab_pages;

  if ((num_tampon != NUM_MOT_IN) && (num_tampon != NUM_MOT_OUT)) {
    return -1;
  }
  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Acces a un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Verifie que le tampon n'est pas deja verrouille
  if (lire_mot_mem(num_tampon+1,num_tab_pages,mot) == -1)
    return -1;
  if ((int)mot[3] == F_LOCK) 
    return -3;
  // Verrouille le tampon
  mot[3] = (octet_t)F_LOCK;
  if (ecrire_mot_mem(num_tampon+1,num_tab_pages,mot) == -1)
    return -1;
  return 0;
}

/**
 * Deverrouille un tampon:
 * - Deverrouille le tampon
 * - Reveille un processus ds l'etat E_ATT_LECT ou E_ATT_ECRIT
 * \param pid : num du processus
 * \param num_p_tampon : num du mot ds la table des pages du processus correspondant au tampon
 * \return 
 * - -1 en cas d'erreur
 * - 0  sinon
 */
int deverrouiller_tampon (int pid, int num_tampon) {
  mot_t mot;
  int num_tab_pages;
  int i;
  int etat;

  if ((num_tampon != NUM_MOT_IN) && (num_tampon != NUM_MOT_OUT)) {
    return -1;
  }
  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Acces a un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Deverrouille le tampon
  mot[3] = (octet_t)F_UNLOCK;
  if (ecrire_mot_mem(num_tampon+1,num_tab_pages,mot) == -1)
    return -1;
  // Reveille le processus pid s'il etait en attente de lecture / ecriture
  switch (num_tampon) {
  case NUM_MOT_IN:
    etat = E_ATT_LECT;
    break;
  case NUM_MOT_OUT:
    etat = E_ATT_ECRIT;
    break;
  }
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) != -1) {
    if (lire_etat(mot) == etat) {
      // Reveille le processus i
      liste_add_pid(liste_proc,pid);
      // MAJ la table des processus
      changer_etat(E_ELIGIBLE,mot);
      if (ecrire_mot_mem(i,NUM_PAGE_META,mot) != -1)
        return 0;
      liste_supp_pid(liste_proc,i);
    }
  }
  return 0;
}

/**
 * Fonction qui ecrit un caractere dans le tampon d'entree d'un processus.\n
 * Elle est non bloquante.\n
 * A la fin, reveille le proc pid s'il etait en E_ATT_LECT.
 * \param pid : num du processus
 * \param c : le caractere a ecrire
 * \return 
 * - -1 en cas d'erreur inconnue
 * - -2 plus de place dans le tampon
 * - -3 tampon verrouille
 * - 0  sinon
 */
int ecrire_stdin (int pid, char c) {
  mot_t mot,mot_tampon;
  int num_tab_pages,num_p_tampon;
  int deb, longueur,fin;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Tentative d'ecriture dans une page d'un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Verrouille le tampon
  switch (verrouiller_tampon(pid,NUM_MOT_IN)) {
  case -1:
    return -1;
  case -3:
    return -3;
  }
  // Cherche la page contenant le tampon d'entree
  if (lire_mot_mem(NUM_MOT_IN,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  switch ((int)mot[2]) {
  case PP_NULL:
    // On n'a encore jamais rien ecrit ds le tampon
    if ((num_p_tampon=page_forcee(pid)) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_IN);
      return -1;
    }
    // Met a jour la table des pages
    mot[0] = 0;
    mot[1] = 0;
    mot[2] = PP_RAM;
    mot[3] = num_p_tampon;
    if (ecrire_mot_mem(NUM_MOT_IN,num_tab_pages,mot) == -1) {
      deverrouiller_tampon(pid,NUM_MOT_IN);
      liberer_page(num_p_tampon);
      return -1;
    }
    break;
  case PP_SWAP:
    //On rapatrie le tampon en mem RAM
    if ((num_p_tampon=rapatrier_page_swap(pid,(int)mot[3])) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_IN);
      return -1;
    }
    break;
  case PP_RAM:
    num_p_tampon = (int)mot[3];
    break;
  default:
    fprintf(stderr,"Table des pages %d du processus %d incoherente\n",num_tab_pages,pid);
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  deb = (int)mot[0];
  longueur = (int)mot[1];
  // Teste s'il reste de la place
  if (longueur == T_MAX_TAMPON) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -2;
  }
  fin = (deb + longueur) % T_MAX_TAMPON;
  // On ecrit le caractere
  if (lire_mot_mem(fin/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  mot_tampon[fin%T_MOT] = (octet_t)c;
  if (ecrire_mot_mem(fin/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  // MAJ de la table des pages
  mot[1] = (octet_t)longueur+1;
  if (ecrire_mot_mem(NUM_MOT_IN,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  // Deverrouille le tampon
  if (deverrouiller_tampon(pid,NUM_MOT_IN) == -1)
    return -1;
  // Reveille le proc pid s'il etait en E_ATT_LECT.
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) != -1) {
    if (lire_etat(mot) == E_ATT_LECT) {
      // Reveille le processus i
      liste_add_pid(liste_proc,pid);
      // MAJ la table des processus
      changer_etat(E_ELIGIBLE,mot);
      if (ecrire_mot_mem(pid,NUM_PAGE_META,mot) != -1)
        return 0;
      liste_supp_pid(liste_proc,pid);
    }
  }
  return 0;
}

/**
 * Fonction qui lit un caractere dans le tampon d'entree d'un processus.\n
 * La fonction n'est pas bloquante mais elle retourne un code d'erreur qui permet
 * a la fonction appelante de bloquer le processus (E_ATT_LECT). 
 * \param pid : num du processus
 * \param c : le caractere a lire
 * \return 
 * - -1 en cas d'erreur inconnue
 * - -2 rien a lire
 * - -3 tampon verrouille
 * - 0  sinon
 */
int lire_stdin (int pid, char* c) {
  mot_t mot,mot_tampon;
  int num_tab_pages,num_p_tampon;
  int deb, longueur;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Tentative d'ecriture dans une page d'un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Verrouille le tampon
  switch (verrouiller_tampon(pid,NUM_MOT_IN)) {
  case -1:
    return -1;
  case -3:
    return -3;
  }
  // Cherche la page contenant le tampon d'entree
  if (lire_mot_mem(NUM_MOT_IN,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  switch ((int)mot[2]) {
  case PP_NULL:
    // On n'a encore jamais rien ecrit ds le tampon
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -2;
  case PP_SWAP:
    //On rapatrie le tampon en mem RAM
    if ((num_p_tampon=rapatrier_page_swap(pid,(int)mot[3])) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_IN);
      return -1;
    }
    break;
  case PP_RAM:
    num_p_tampon = (int)mot[3];
    break;
  default:
    fprintf(stderr,"Table des pages %d du processus %d incoherente\n",num_tab_pages,pid);
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  deb = (int)mot[0];
  longueur = (int)mot[1];
  // Teste s'il y a quelque chose a lire
  if (longueur == 0) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -2;
  }
  // On lit le caractere
  if (lire_mot_mem(deb/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  *c = (char)mot_tampon[deb%T_MOT];
  // MAJ de la table des pages
  mot[0] = (octet_t)deb+1;
  mot[1] = (octet_t)longueur-1;
  if (ecrire_mot_mem(NUM_MOT_IN,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_IN);
    return -1;
  }
  // Deverrouille le tampon
  if (deverrouiller_tampon(pid,NUM_MOT_IN) == -1)
    return -1;
  return 0;
}

/**
 * Fonction qui ecrit un caractere dans le tampon d'entree d'un processus.\n
 * La fonction n'est pas bloquante mais elle retourne un code d'erreur qui permet
 * a la fonction appelante de bloquer le processus (E_ATT_ECRIT). 
 * \param pid : num du processus
 * \param c : le caractere a ecrire
 * \return 
 * - -1 en cas d'erreur inconnue
 * - -2 plus de place dans le tampon
 * - -3 tampon verrouille
 * - 0  sinon
 */
int ecrire_stdout (int pid, char c) {
  mot_t mot,mot_tampon;
  int num_tab_pages,num_p_tampon;
  int deb, longueur,fin;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Tentative d'ecriture dans une page d'un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Verrouille le tampon
  switch (verrouiller_tampon(pid,NUM_MOT_OUT)) {
  case -1:
    return -1;
  case -3:
    return -3;
  }
  // Cherche la page contenant le tampon d'entree
  if (lire_mot_mem(NUM_MOT_OUT,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  switch ((int)mot[2]) {
  case PP_NULL:
    // On n'a encore jamais rien ecrit ds le tampon
    if ((num_p_tampon=page_forcee(pid)) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_OUT);
      return -1;
    }
    // Met a jour la table des pages
    mot[0] = 0;
    mot[1] = 0;
    mot[2] = PP_RAM;
    mot[3] = num_p_tampon;
    if (ecrire_mot_mem(NUM_MOT_OUT,num_tab_pages,mot) == -1) {
      deverrouiller_tampon(pid,NUM_MOT_OUT);
      liberer_page(num_p_tampon);
      return -1;
    }
    break;
  case PP_SWAP:
    //On rapatrie le tampon en mem RAM
    if ((num_p_tampon=rapatrier_page_swap(pid,(int)mot[3])) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_OUT);
      return -1;
    }
    break;
  case PP_RAM:
    num_p_tampon = (int)mot[3];
    break;
  default:
    fprintf(stderr,"Table des pages %d du processus %d incoherente\n",num_tab_pages,pid);
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  deb = (int)mot[0];
  longueur = (int)mot[1];
  // Teste s'il reste de la place
  if (longueur == T_MAX_TAMPON) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -2;
  }
  fin = (deb + longueur) % T_MAX_TAMPON;
  // On ecrit le caractere
  if (lire_mot_mem(fin/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  mot_tampon[fin%T_MOT] = (octet_t)c;
  if (ecrire_mot_mem(fin/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  // MAJ de la table des pages
  mot[1] = (octet_t)longueur+1;
  if (ecrire_mot_mem(NUM_MOT_OUT,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  // Deverrouille le tampon
  if (deverrouiller_tampon(pid,NUM_MOT_OUT) == -1)
    return -1;
  return 0;
}

/**
 * Fonction qui lit un caractere dans le tampon de sortie d'un processus.\n
 * Elle est non bloquante.\n
 * Reveille le proc pid s'il etait en E_ATT_ECRIT
 * \param pid : num du processus
 * \param c : le caractere a lire
 * \return 
 * - -1 en cas d'erreur inconnue
 * - -2 rien a lire
 * - -3 tampon verrouille
 * - 0  sinon
 */
int lire_stdout (int pid, char* c) {
  mot_t mot,mot_tampon;
  int num_tab_pages,num_p_tampon;
  int deb, longueur;

  // Cherche la table des pages du processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1)
    return -1;
  if (lire_etat(mot) == E_MORT) {
    fprintf(stderr,"Tentative d'ecriture dans une page d'un processus mort (%d)\n",pid);
    return -1;
  }
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_tab_pages = mot[3];
    break;
  case PP_SWAP:
    // Ramene la table des pages du proc en mem RAM
    if ((num_tab_pages=rapatrier_page_swap(pid,(int)mot[3])) < 0)
      return -1;
    break;
  }
  // Verrouille le tampon
  switch (verrouiller_tampon(pid,NUM_MOT_OUT)) {
  case -1:
    return -1;
  case -3:
    return -3;
  }
  // Cherche la page contenant le tampon d'entree
  if (lire_mot_mem(NUM_MOT_OUT,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  switch ((int)mot[2]) {
  case PP_NULL:
    // On n'a encore jamais rien ecrit ds le tampon
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -2;
  case PP_SWAP:
    //On rapatrie le tampon en mem RAM
    if ((num_p_tampon=rapatrier_page_swap(pid,(int)mot[3])) < 0) {
      deverrouiller_tampon(pid,NUM_MOT_OUT);
      return -1;
    }
    break;
  case PP_RAM:
    num_p_tampon = (int)mot[3];
    break;
  default:
    fprintf(stderr,"Table des pages %d du processus %d incoherente\n",num_tab_pages,pid);
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  deb = (int)mot[0];
  longueur = (int)mot[1];
  // Teste s'il y a quelque chose a lire
  if (longueur == 0) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -2;
  }
  // On lit le caractere
  if (lire_mot_mem(deb/T_MOT,num_p_tampon,mot_tampon) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  *c = (char)mot_tampon[deb%T_MOT];
  // MAJ de la table des pages
  mot[0] = (octet_t)deb+1;
  mot[1] = (octet_t)longueur-1;
  if (ecrire_mot_mem(NUM_MOT_OUT,num_tab_pages,mot) == -1) {
    deverrouiller_tampon(pid,NUM_MOT_OUT);
    return -1;
  }
  // Deverrouille le tampon
  if (deverrouiller_tampon(pid,NUM_MOT_OUT) == -1)
    return -1;
  // Reveille le proc pid s'il etait en E_ATT_ECRIT
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) != -1) {
    if (lire_etat(mot) == E_ATT_ECRIT) {
      // Reveille le processus i
      liste_add_pid(liste_proc,pid);
      // MAJ la table des processus
      changer_etat(E_ELIGIBLE,mot);
      if (ecrire_mot_mem(pid,NUM_PAGE_META,mot) != -1)
        return 0;
      liste_supp_pid(liste_proc,pid);
    }
  }
  return 0;
}

/**
 * Fonction qui cree un fichier contenant l'image de toute la memoire du systeme
 * \param pid : num du processus qui a fait appel a la fonction\n
 *              (permet de donner un nom au fichier image)
 * \return rien
 */
void creer_image_totale (int pid) {
  char *nom, *comment;
  int desc;
  page_t *pp;

  // Nom du fichier
  nom = malloc((1+2+5)*sizeof(char));
  sprintf(nom,"%d.image",pid);
  // Ouverture du fichier
  if ((desc=open(nom,O_CREAT|O_TRUNC|O_WRONLY,0666)) == -1) {
    fprintf(stderr,"Impossible d'ouvrir le fichier %s pour image de la mem\n",nom);
    return;
  }
  fprintf(stderr,"Creation de l'image de toutes la mem du sys\n");
  // La mem principale
  strcpy(comment,"--Image de la mem principale--");
  if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  if (write(desc,mem,NPP*sizeof(page_t)) != NPP*sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  // La mem secondaire
  strcpy(comment,"--Image de la mem secondaire--");
  if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  if (lseek(swap_d,0,SEEK_SET) == -1) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  if (read(swap_d,pp,NPPS*sizeof(page_t)) != NPPS*sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  if (write(desc,pp,NPPS*sizeof(page_t)) != NPPS*sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  return;
}

/**
 * Fonction qui cree un fichier contenant l'image du processus a un instant donne.\n
 * Cette image represente:
 * - la table des processus\n
 *   (si le processus pid n'est pas lisible ou est mort, toute la mem est recopiee)
 * - la table des pages du processus
 * - toutes les pages associes (code, pile, donnee et tampons E/S)
 * \param pid : num du processus
 * \return rien
 */
void creer_image_mem (int pid) {
  mot_t mot;
  char *nom, *comment;
  int desc;
  mot_t *pp, *ptp;
  int num_pp;
  int i,j;
  int num_mot;

  // Nom du fichier
  nom = malloc((1+2+5)*sizeof(char));
  sprintf(nom,"%d.image",pid);
  // Ouverture du fichier
  if ((desc=open(nom,O_CREAT|O_TRUNC|O_WRONLY,0666)) == -1) {
    fprintf(stderr,"Impossible d'ouvrir le fichier %s pour image de la mem\n",nom);
    return;
  }
  // Lecture de la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot) == -1) {
    creer_image_totale(pid);
    return;
  }
  if (lire_etat(mot) == E_MORT) {
    creer_image_totale(pid);
    return;
  }
  // Copie de la table des processus
  comment = malloc(sizeof(page_t));
  strcpy(comment,"--Table des processus--");
  if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  if (write(desc,mem,sizeof(page_t)) != sizeof(page_t)) {
    fprintf(stderr,"Image de la mem non terminee\n");
    close(desc);
    return;
  }
  // Copie de la table des pages
  switch (lire_type_pp(mot)) {
  case PP_RAM:
    num_pp = (int)mot[3];
    ptp = mem[num_pp];
    if (write(desc,ptp,sizeof(page_t)) != sizeof(page_t)) {
      fprintf(stderr,"Image de la mem non terminee\n");
      close(desc);
      return;
    }
    break;
  case PP_SWAP:
    num_pp = (int)mot[3];
    if (lseek(swap_d,num_pp*sizeof(page_t),SEEK_SET) == -1) {
      fprintf(stderr,"Image de la mem non terminee\n");
      close(desc);
      return;
    }
    if (read(swap_d,ptp,sizeof(page_t)) != sizeof(page_t)) {
      fprintf(stderr,"Image de la mem non terminee\n");
      close(desc);
      return;
    }
    if (write(desc,ptp,sizeof(page_t)) != sizeof(page_t)) {
      fprintf(stderr,"Image de la mem non terminee\n");
      close(desc);
      return;
    }
    break;
  }
  // Copie des pages de code, donnees et pile
  for (i=0 ; i<3 ; i++) {
    for (j=T_MAX_ZONES*i ; j<T_MAX_ZONES*(i+1) ; j++) {
      if (ptp[j][2] == PP_NULL)
        break;
      else if (ptp[j][2] == PP_RAM) {
        num_pp = ptp[j][3];
        sprintf(comment,"Page %d en RAM\n",num_pp);
        if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
        pp = mem[num_pp];
        if (write(desc,pp,sizeof(page_t)) != sizeof(page_t)) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
      }
      else if (ptp[j][2] == PP_SWAP) {
        num_pp = ptp[j][3];
        sprintf(comment,"Page %d en SWAP\n",num_pp);
        if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
        if (lseek(swap_d,num_pp*sizeof(page_t),SEEK_SET) == -1) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
        if (read(swap_d,pp,sizeof(page_t)) != sizeof(page_t)) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
        if (write(desc,pp,sizeof(page_t)) != sizeof(page_t)) {
          fprintf(stderr,"Image de la mem non terminee\n");
          close(desc);
          return;
        }
      }
    }
  }
  // Copie des tampons d'E/S
  for (i=0 ; i<2 ; i++) {
    if (i == 0)
      num_mot = NUM_MOT_IN;
    else
      num_mot = NUM_MOT_OUT;
    if (ptp[num_mot][2] == PP_NULL)
      break;
    else if (ptp[num_mot][2] == PP_RAM) {
      num_pp = ptp[num_mot][3];
      sprintf(comment,"Page %d en RAM\n",num_pp);
      if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
      pp = mem[num_pp];
      if (write(desc,pp,sizeof(page_t)) != sizeof(page_t)) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
    }
    else if (ptp[num_mot][2] == PP_SWAP) {
      num_pp = ptp[num_mot][3];
      sprintf(comment,"Page %d en SWAP\n",num_pp);
      if (write(desc,comment,sizeof(page_t)) != sizeof(page_t)) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
      if (lseek(swap_d,num_pp*sizeof(page_t),SEEK_SET) == -1) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
      if (read(swap_d,pp,sizeof(page_t)) != sizeof(page_t)) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
      if (write(desc,pp,sizeof(page_t)) != sizeof(page_t)) {
        fprintf(stderr,"Image de la mem non terminee\n");
        close(desc);
        return;
      }
    }
  }
}


/**
 * Fonction qui interprete les signaux pendant d'un processus
 * \param pid : num du processus
 * \return rien
 */
void interprete_signaux (int pid) {
  signal_t *sig;
  mot_t mot_proc,mot_pile;
  vad_t adv_pile;

  mot_pile[0] = (octet_t)M_UT;
  adv_pile.num = 0;
  adv_pile.zone = Z_PILE;
  // Verifie que le num de processus est valide
  if (pid < 0 || pid > N_MAX_PROC) {
    fprintf(stderr, "Le proc %d n'existe pas\n", pid);
    return;
  }
  // Lit l'entree du processus ds la table des processus
  if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
    return;
  // Regarde si le processus est toujours actif
  if (lire_etat(mot_proc) == E_MORT)
    return;
  // Regarde si le processus a des signaux pendants
  if (SIFUNSIG(mot_proc[0]))
    return;
  // Regarde quels signaux sont pendants
  if (SIFKILL(mot_proc[0])){
    printf("Mort du proc %d par Arme fatale\n",pid);
    if (terminer_proc(pid) != 0)
      fprintf(stderr,"Processus %d non termine\n",pid);
    return;
  }
  if (SIFMEM(mot_proc[0])) {
    printf("Violation de memoire au processus %d ", pid);
    // Regarde si le signal est capture par le proc
    if ((sig=rechercher_sig(pid,S_MEM)) == NULL) {
      // Comportement par defaut
      printf("comportement par defaut\n");
      // Cree une image mem
      creer_image_mem(pid);
      // Termine le proc
      if (terminer_proc(pid) != 0)
        fprintf(stderr,"Processus %d non termine\n",pid);
      return;
    }
    else {
      // Le signal est capture
      // Incremente de 2 la pile pour y mettre:
      // P1: le signal capture
      /* Je ne sais pas comment faire pour supprimer le param en sortant du handler
      changer_esp(pid,1);
      mot_pile[3] = (octet_t)sig_lib2ksinu(S_MEM);
      if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
        fprintf(stderr,"Capture de S_MEM echouee\n");
      else {
      */
        // P0: le compteur programme de retour (ie actuel)
        changer_esp(pid,1);
        mot_pile[3] = mot_proc[1];
        if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
          fprintf(stderr,"Capture de S_MEM echouee\n");
        else {
          // Change le compteur programme
          mot_proc[1] = (octet_t)sig->cp;
          if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
            fprintf(stderr,"Capture de S_MEM echouee\n");
        }
    }
  }
  if (SIFSUSP(mot_proc[0])) {
    printf("Suspension du processus %d\n",pid);
    // Regarde si le signal est ignore par le proc
    if ((sig=rechercher_sig(pid,S_SUSP)) == NULL) {
      // Comportement par defaut
      printf("comportement par defaut\n");
      // Suspension de l'execution
      liste_supp_pid(liste_proc,pid);
      changer_etat(E_SUSP,mot_proc);
      if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
        fprintf(stderr,"Suspension du proc %d echouee\n",pid);
        liste_add_pid(liste_proc,pid);
      }
    }
  }
  if (SIFREP(mot_proc[0])) {
    printf("Reprise du processus %d\n", pid);
    if (lire_etat(mot_proc) == E_SUSP) {
      liste_add_pid(liste_proc,pid);
      changer_etat(E_ELIGIBLE,mot_proc);
      if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
        fprintf(stderr,"Reprise du proc %d echouee\n",pid);
        liste_supp_pid(liste_proc,pid);
      }
    }
  }
  if (SIFUSR1(mot_proc[0])) {
    printf("Signal non definie 1 recu par processus %d\n",pid);
    // Regarde si le signal est capture ou ignore par le proc
    if ((sig=rechercher_sig(pid,S_USR1)) == NULL) {
      // Comportement par defaut
      printf("comportement par defaut\n");
      // Termine le proc
      if (terminer_proc(pid) != 0)
        fprintf(stderr,"Processus %d non termine\n",pid);
      return;
    }
    else if (sig->cp != 0) {
      // Le signal est capture
      // Incremente de 2 la pile pour y mettre:
      // P1: le signal capture
      /*
      changer_esp(pid,1);
      mot_pile[3] = (octet_t)sig_lib2ksinu(S_USR1);
      if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
        fprintf(stderr,"Capture de S_USR1 echouee\n");
      else {
      */
        // P0: le compteur programme de retour (ie actuel)
        changer_esp(pid,1);
        mot_pile[3] = mot_proc[1];
        if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
          fprintf(stderr,"Capture de S_USR1 echouee\n");
        else {
          // Change le compteur programme
          mot_proc[1] = (octet_t)sig->cp;
          if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
            fprintf(stderr,"Capture de S_USR1 echouee\n");
        }
    }
  }
  if (SIFUSR2(mot_proc[0])) {
    printf("Signal non definie 2 recu par processus %d\n",pid);
    // Regarde si le signal est capture par le proc
    if ((sig=rechercher_sig(pid,S_USR2)) == NULL) {
      // Comportement par defaut
      printf("comportement par defaut\n");
      // Cree une image mem
      creer_image_mem(pid);
      // Termine le proc
      if (terminer_proc(pid) != 0)
        fprintf(stderr,"Processus %d non termine\n",pid);
      return;
    }
    else if (sig->cp != 0) {
      // Le signal est capture
      // Incremente de 2 la pile pour y mettre:
      // P1: le signal capture
      /*
      changer_esp(pid,1);
      mot_pile[3] = (octet_t)sig_lib2ksinu(S_USR2);
      if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
        fprintf(stderr,"Capture de S_USR2 echouee\n");
      else {
      */
        // P0: le compteur programme de retour (ie actuel)
        changer_esp(pid,1);
        mot_pile[3] = mot_proc[1];
        if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
          fprintf(stderr,"Capture de S_USR2 echouee\n");
        else {
          // Change le compteur programme
          mot_proc[1] = (octet_t)sig->cp;
          if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
            fprintf(stderr,"Capture de S_USR2 echouee\n");
        }
    }
  }
  if (SIFINSTI(mot_proc[0])) {
    printf("Instruction illegale ds le processus %d\n",pid);
    // Regarde si le signal est capture par le proc
    if ((sig=rechercher_sig(pid,S_INSTI)) == NULL) {
      // Comportement par defaut
      printf("comportement par defaut\n");
      // Cree une image mem
      creer_image_mem(pid);
      // Termine le proc
      if (terminer_proc(pid) != 0)
        fprintf(stderr,"Processus %d non termine\n",pid);
      return;
    }
    else {
      // Le signal est capture
      // Incremente de 2 la pile pour y mettre:
      // P1: le signal capture
      /*
      changer_esp(pid,1);
      mot_pile[3] = (octet_t)sig_lib2ksinu(S_INSTI);
      if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
        fprintf(stderr,"Capture de S_INSTI echouee\n");
      else {
      */
        // P0: le compteur programme de retour (ie actuel)
        changer_esp(pid,1);
        mot_pile[3] = mot_proc[1];
        if (ecrire_mot(pid,adv_pile,mot_pile) < 0)
          fprintf(stderr,"Capture de S_INSTI echouee\n");
        else {
          // Change le compteur programme
          mot_proc[1] = (octet_t)sig->cp;
          if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
            fprintf(stderr,"Capture de S_INSTI echouee\n");
        }
    }
  }
}

/**
 * Fonction qui execute l'instruction courante du processus donne en param
 * \param pid : num du processus
 * \return rien
 */
void next_inst (int pid) {
  mot_t mot,mot_proc;
  vad_t ad_v;
  int ret=0;

  // Verifie que le num de processus est valide
  if (pid < 0 || pid > N_MAX_PROC) {
    fprintf(stderr, "Le proc %d n'existe pas\n", pid);
    return;
  }
  else {
    if (lire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1)
      return;
    if (lire_etat(mot_proc) == E_MORT) {
      fprintf(stderr, "Tentative d'execution d'un proc termine (%d est vide)\n", pid);
      return;
    }
  }
  // Lit l'instruction courante
  ad_v.zone = Z_CODE;
  ad_v.num = (int)mot_proc[1];
  switch (lire_mot(pid, ad_v, mot)) {
  case -1:
    ret = -1;
    break;
  case -2:
    ret = -2;
    break;
  default:
    switch ((int)mot[0]) {
    case I_AFFECT:
      ret = cmd_affect(pid,mot);
      break;
    case I_AFFECTP:
      ret = cmd_affectp(pid,mot);
      break;
    case I_SI:
      ret = cmd_si(pid,mot);
      break;
    case I_SAUT:
      ret = cmd_saut(pid,mot);
      break;
    case I_CPILE:
      ret = cmd_cpile(pid,mot);
      break;
    case I_DPILE:
      ret = cmd_dpile(pid,mot);
      break;
    case I_APPEL:
      ret = cmd_appel(pid,mot);
      break;
    case I_RETOUR:
      ret = cmd_retour(pid);
      break;
    case I_TEST:
      ret = cmd_test(pid,mot);
      break;
    case I_TRAPPE:
      switch ((int)mot[2]) {
      case T_CLONE:
        ret = cmd_clone(pid);
        break;
      case T_RECOUVRE:
        ret = cmd_recouvre(pid);
        break;
      case T_ATTENDS:
        ret = cmd_attends(pid);
        break;
      case T_FIN:
        ret = cmd_fin(pid);
        break;
      case T_CAPTURE:
        ret = cmd_capture(pid);
        break;
      case T_EMETS:
        ret = cmd_emets(pid);
        break;
      case T_ID:
        ret = cmd_id(pid);
        break;
      case T_IDP:
        ret = cmd_idp(pid);
        break;
      case T_LIT:
        ret = cmd_lit(pid);
        break;
      case T_ECRIT:
        ret = cmd_ecrit(pid);
        break;
      default:
        ret = -1;
      }
      break;
    case I_AFFECTSP:
      ret = cmd_affectsp(pid,mot);
      break;
    case I_DONNEES:
      ret = cmd_donnees(pid,mot);
      break;
    default:
      ret = -1;
    }
  }
  switch (ret) {
  case -1:
    fprintf(stderr, "Instructioin illegale du proc %d\n",pid);
    mot_proc[0] = (octet_t)mot_proc[0] | S_INSTI;
    if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
      // Termine le processus
      terminer_proc(pid);
      return;
    }
    break;
  case -2:
    fprintf(stderr, "Violation de memoire dans le proc %d\n",pid);
    mot_proc[0] = (octet_t)mot_proc[0] | S_MEM;
    if (ecrire_mot_mem(pid,NUM_PAGE_META,mot_proc) == -1) {
      // Termine le processus
      terminer_proc(pid);
      return;
    }
    break;
  }
}
