/**
  \file interpreteur.h
  \brief Entete de interpreteur.c
*/

#ifndef INTERPRETEUR_H
#define INTERPRETEUR_H

#include "lib_ksinu.h"
#include "fifo.h"


/**************************************************************************************
 Entierement fait par Martin Ludovic
**************************************************************************************/


/*******************************************************************************
 Gestion (liberation / reservation) des num de processus et des pages physiques
********************************************************************************/

/// Fonction permettant de liberer un num de processus
int liberer_pid (int pid);

/// Fonction retournant un num de processus libre
int pid_libre();

/// Fonction permettant de liberer un num de page physique ds la memoire
int liberer_page (int page);

/// Fonction retournant un num de page physique libre ds la memoire
int page_libre (int pid);

/// Fonction retournant un num de page physique en memoire en l'ayant eventuellement liberee de force
int page_forcee (int pid);

/// Fonction permettant de liberer un num de page physique ds le swap
int liberer_page_swap (int page);

/// Fonction retournant un num de page physique libre ds le swap
int page_swap_libre();


/**************************************************************
 Interaction entre memoire principale et secondaire
**************************************************************/

/// Fonction permettant de mettre en mem principale une page se trouvant ds le swap
int rapatrier_page_swap (int pid, int page);

/// Fonction permettant de copier une page du swap ds la mem princale
int page_swap2mem (int num_page_swap, int num_page_mem);

/// Fonction permettant de copier une page de la mem princale ds le swap
int page_mem2swap (int num_page_mem, int num_page_swap);


/***************************************************************
 Gestion de la table des pages des processus
****************************************************************/

// Constantes indiquant la zone de la memoire du precessus
#define Z_CODE 0     // Zone de code
#define Z_DONNEES 1  // Zone de donnees
#define Z_PILE 2     // Zone de pile

/// Structure definissant une adresse virtuelle
typedef struct {
  int zone; ///< Zone mmoire du processus (code, pile ou donnees)
  int num;  ///< Numero du mot ds la zone
} vad_t;

/// Structure definissant un mot physique
typedef struct {
  int type;   ///< Type de la mem ou se trouve la page (RAM ou swap)
  int page;   ///< Numero de la page
  int mot;    ///< Numero du mot
} info_mot_t;

/// Fonction retournant le num de page (ds la mem ou le swap) d'une adresse virtuelle
int adresse_phys (int pid, vad_t ad_v, info_mot_t *info);

/// Fonction permettant d'ajouter une page a la table des pages d'un processus
int page_add2proc (int pid, int zone, int page);

/// Fonction pour supprimer la derniere page d'une zone de la table des pages d'un proc
int page_supp2proc (int pid, int zone);

/// Fonction permettant de modifier le pid du processus pere
int changer_ppid (int pid, int ppid);

/// Fonction permettant de lire le pid du processus pere
int lire_ppid (int pid);

/// Modifie la hauteur de la pile
int changer_esp (int pid, int op);

/// Retourne la hauteur de la pile
int lire_esp (int pid);


/**********************************************************************
 Acces (lecture / ecriture) a la memoire (principale et secondaire)
***********************************************************************/

/// Fonction permettant de lire un mot se trouvant ds la mem principale
int lire_mot_mem (int num_mot, int num_pp, mot_t mot);

/// Fonction permettant d'ecrire un mot se trouvant ds la mem principale
int ecrire_mot_mem (int num_mot, int num_pp, mot_t mot);

/// Fonction permettant de lire un mot a partir d'une adresse virtuelle
int lire_mot (int pid, vad_t ad_v, mot_t mot);

/// Fonction qui ecrit un mot a partir d'une adresse virtuelle
int ecrire_mot (int pid, vad_t ad_v, mot_t mot);

/// Fonction permettant d'ecrire une page dans la mem swap
int ecrire_page_swap (page_t page, int num_page);


/**************************************************************
 Interpretation d'une entree de la table des processus
**************************************************************/

/// Fonction qui permet de changer l'etat d'un processus
int changer_etat(int etat, mot_t mot);

/// Fonction qui permet de lire l'etat d'un processus
int lire_etat(mot_t mot);

/// Fonction qui permet de changer le type de la table des pages d'un processus
int changer_type_pp(int type, mot_t mot);

/// Fonction qui permet de lire le type de la table des pages d'un processus
int lire_type_pp(mot_t mot);


/**************************************************************
 Acces au tampons d'entrees / sorties d'un processus
**************************************************************/

/// Verrouille un tampon
int verrouiller_tampon (int pid, int num_tampon);

/// Deverrouille un tampon
int deverrouiller_tampon (int pid, int num_tampon);

/// Fonction qui ecrit un caractere dans le tampon d'entree d'un processus
int ecrire_stdin (int pid, char c);

/// Fonction qui lit un caractere dans le tampon d'entree d'un processus
int lire_stdin (int pid, char* c);

/// Fonction qui ecrit un caractere dans le tampon de sortie d'un processus
int ecrire_stdout (int pid, char c);

/// Fonction qui lit un caractere dans le tampon de sortie d'un processus
int lire_stdout (int pid, char* c);


/**************************************************************
 Interpretation d'une instruction
**************************************************************/

// Code des instructions
#define I_AFFECT 1
#define I_AFFECTP 2
#define I_SI 3
#define I_SAUT 4
#define I_CPILE 5
#define I_DPILE 6
#define I_APPEL 7
#define I_RETOUR 8
#define I_TEST 9
#define I_TRAPPE 10
#define I_AFFECTSP 11
#define I_DONNEES 255

// Code des trappes
#define T_CLONE 1
#define T_RECOUVRE 2
#define T_ATTENDS 3
#define T_FIN 4
#define T_CAPTURE 5
#define T_EMETS 6
#define T_ID 7
#define T_IDP 8
#define T_LIT 9
#define T_ECRIT 10

/// Fonction qui execute l'instruction courante du processus donne
void next_inst (int pid);

/// Fonction qui interprete les signaux pendant d'un processus
void interprete_signaux (int pid);

/// Fonction qui cree un fichier contenant l'image du proc a un instant donne
void creer_immage_mem (int pid);

/// Fonction qui cree un fichier contenant l'image de toute la mem du sys
void creer_image_totale (int pid);

#endif
