/**
 * @file lib_ksinu.h
 * Bibliotheque standard pour le projet KSInu. Definit des constantes, structures
 * et types pour gerer :
 * - la memoire principale et secondaire
 * - les processus
 */

#ifndef LIB_KSINU_H
#define LIB_KSINU_H

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>


/***************************************************
 Definition de la memoire pour la machine virtuelle
***************************************************/

// Definition de la memoire
#define T_MOT 4                    // Taille en octets d'un mot
#define NPP 16                     // Nb de pages physiques en memoire
#define NMPPP 32                   // Nb max de mot dans une page
#define NPPS 255                   // Nb de pages physiques dans le swap
#define T_PAGE (NMPPP*T_MOT)       // Taille en octets d'une page
#define T_MAX_TAMPON (NMPPP*T_MOT) // Taille max d'un tampon (en octet)

// Type pour manipuler la memoire
typedef unsigned char octet_t; // octet_t est un octet non signe
typedef octet_t mot_t[T_MOT];  // mot_t represente un mot mem ie une instruction
typedef mot_t page_t[NMPPP];   // page_t represente un page physique

// Organisation de la zone memoire d'un processus (table des pages d'un processus)
#define T_MAX_ZONES 8                  // Taille max (en mots) des zones de code, donnees et pile
#define NUM_MOT_IN (T_MAX_ZONES*3)     // Num du mot pointant sur le tampon d'entree
#define NUM_MOT_OUT (T_MAX_ZONES*3+2)  // Num du mot pointant sur le tampon de sortie
#define NUM_MOT_ESP (T_MAX_ZONES*3+4)  // Num du mot contenant la hauteur de la pile
#define SP_MAX (T_MAX_ZONES*NMPPP)     // Hauteur maximale de la pile
#define NUM_MOT_PPID (T_MAX_ZONES*3+5) // Num du mot contenant le ppid du processus
#define NO_PERE 0                      // ppid d'un processus orphelin (qui n'a pas de pere)
#define NUM_MOT_EXIT (T_MAX_ZONES*3+6) // Num du mot contenant la valeur de retour du proc

// Verrous
#define F_UNLOCK 0
#define F_LOCK 1

// Variables globales simulant les memoires
page_t mem[NPP]; // Tableau simulant la memoire principale (RAM)
int swap_d;      // Descripteur du fichier simulant la memoire secondaire (swap)



/***************************************************
 Definition pour la table des processus
***************************************************/

// Constantes de la table des processus
#define N_MAX_PROC (NMPPP-NPP/T_MOT-1) // Taille de la table des processus
#define DEBUT_T_PAGES (N_MAX_PROC+1)   // Debut de la table des pages ds la page des meta-info
#define DEBUT_T_PROC 1   // Debut de la table des processus ds la page des meta-info
#define NUM_PAGE_META 0  // Numero de la page des meta-inof contenant notamment la table des processus
#define NUM_PAGE_SWAP 1  // Num de la page mem contenant la 1ere page de la liste des pages swap libres
#define PAGE_VIDE 0      // Indique que l'emplacement (numero de pp) dans la memoire est libre
#define NO_MORE_PID 0    // Plus d'emplacement libre ds la table des proc
#define NO_MORE_PAGE 0   // Plus de page physique libre en memoire
#define NO_MORE_SWAP 255 // Plus de page physique libre ds le swap

// Constantes indiquant le type de la page physique
#define PP_RAM 1   // RAM
#define PP_SWAP 2  // Swap
#define PP_NULL 0  // Inutilise par le processus

// Constantes indiquant l'etat du processus
#define E_MORT 0      // Processus mort
#define E_ELIGIBLE 1  // Processus eligible
#define E_ZOMBI 2     // Processus zombi
#define E_SUSP 3      // Processus suspendu
#define E_ATT_FILS 4  // Processus attendant la mort d'un fils
#define E_ATT_LECT 5  // Processus attendant de pouvoir lire son entree standard
#define E_ATT_ECRIT 6 // Processus attendant de pouvoir ecrire sur sa sortie standard

// Constantes indiquant l'etat du drapeau de saut conditionnel
#define C_VRAI 1 // La condition est vraie
#define C_FAUX 0 // La condition est fausse

// Constantes indiquant l'etat d'un mot dans une page
#define M_NU 0 // Mot non utilise (par ex ceux qui se trouvent apres le haut de la pile)
#define M_UT 1 // Mot utilise



/************************************************
 Definitions pour les signaux
************************************************/

// Les signaux
#define S_NULL 0   // Aucun signal
#define S_KILL 1   // Arme fatale
#define S_MEM 2    // Violation de memoire
#define S_SUSP 4   // Suspension
#define S_REP 8    // Reprise
#define S_USR1 16  // Non definie 1
#define S_USR2 32  // Non definie 2
#define S_INSTI 64 // Instruction illegale

// Macros associees aux signaux
/// Vrai si aucun signal n'est attache au processus
#define SIFUNSIG(octet) ((int)octet == S_NULL)
/// Vrai si S_KILL est pendant
#define SIFKILL(octet) ((octet & S_KILL) == S_KILL)
/// Vrai si S_MEM est pendant
#define SIFMEM(octet) ((octet & S_MEM) == S_MEM)
/// Vrai si S_SUSP est pendant
#define SIFSUSP(octet) ((octet & S_SUSP) == S_SUSP)
/// Vrai si S_REP est pendant
#define SIFREP(octet) ((octet & S_REP) == S_REP)
/// Vrai si S_USR1 est pendant
#define SIFUSR1(octet) ((octet & S_USR1) == S_USR1)
/// Vrai si S_USR2 est pendant
#define SIFUSR2(octet) ((octet & S_USR2) == S_USR2)
/// Vrai si S_INSTI est pendant
#define SIFINSTI(octet) ((octet & S_INSTI) == S_INSTI)

/// Structure representant un signal intercepte
typedef struct signal {
  int signal;          ///< Signal capture ou ignore
  int cp;              ///< Si capture : compteur programme du handler. Si ignore : S_IGNORE 
  struct signal *next; ///< Signal intercepte suivant
} signal_t;

/// Table de capture des signaux
signal_t* table_capture[N_MAX_PROC];


/************************************************
 Implementation d'une pile fifo
************************************************/

/// Structure d'une cellule de la pile.
struct cellule {
  // Champ propre a la cellule
  struct cellule *suivant; ///< Cellule suivante ds la pile.
  // Champs pour les processus
  int pid;                 ///< Numero du processus en cours d'execution
  int temps;               ///< Nb de fois consecutives que le proc a ete appele.
  // Champ pour les pages
  int page;                ///< Numero d'une page physique de la memoire principale
};

/// Le type pile fifo.
typedef struct {
  struct cellule *tete;
} l_fifo;


/***************************************************
 Execution des processus
***************************************************/

/// Nb d'instructions executees a la suite par un processus av de rendre la main a un autre proc
int quantum;

/// Ordonnancement des processus
l_fifo* liste_proc;

/// Gestion de la memoire principale (memorise les acces a chaque page)
l_fifo* liste_pages;

#endif
