/**
 * \file signaux.c
 * Fonctions permettant de manipuler la liste de signaux a intercepter pour 
 * chaque processus.
 */

#include "signaux.h"


/**
 * Ajoute ou modifie un signal a intercepte au processus
 * \param pid : le num du processus a ajouter
 * \param sig : le signal a ajouter ou modifier
 * \return rien
 */
void add_sig (int pid, signal_t *sig) {
  signal_t *p_sig; 
  signal_t *p_tmp; 

  // Regarde si le signal n'est pas deja intercepte
  p_sig = rechercher_sig(pid, sig->signal);
  if (p_sig != NULL) {
    p_sig->cp = sig->cp;
    free (sig);
  }
  else {
    // Le signal n'est pas deja intercepte
    if (table_capture[pid] == NULL) {
      // C'est le premier signal a intercepte
      table_capture[pid] = p_sig;
    }
    else { 
      p_tmp = table_capture[pid];
      while (p_tmp->next) {
        p_tmp = p_tmp->next;
      }
      p_tmp->next = sig;
    }
  }
}

/**
 * Recherche si le processus intercepte le signal
 * \param pid : le num du processus dont on scan la liste des signaux
 * \param sig : le signal rechercher
 * \return 
 * - NULL si le signal n'a pas ete trouve
 * - un pointeur sur lui sinon
 */
signal_t* rechercher_sig (int pid, int sig) {
  signal_t *p_tmp; 

  // Cas de la liste vide
  if (table_capture[pid] == NULL) {
    return NULL;
  }
  p_tmp = table_capture[pid];
  while (p_tmp) {
    if (p_tmp->signal == sig) {
      return p_tmp;
    }
    p_tmp = p_tmp->next;
  }
  return NULL;
}

/**
 * Supprime un signal a intercepter de la liste des signaux du processus
 * \param pid : le num du processus dont on veut supprimer un signal
 * \param sig : le signal a supprimer
 * \return 
 * - -1 en cas de liste des signaux vide
 * - -2 Le signal n'a pas ete trouve
 * -  0 sinon
 */
int supp_sig (int pid, int sig) {
  signal_t *p_tmp; 
  signal_t *p_clr; 

  // Cas de la liste vide
  if (table_capture[pid] == NULL) {
    fprintf(stderr,"Impossible de supprimer un signal a une liste vide\n");
    return -1;
  }
  else {
    if (table_capture[pid]->signal == sig) {
      p_clr = table_capture[pid];
      table_capture[pid] =  table_capture[pid]->next;
      free (p_clr);
      return 0;
    }
    p_tmp = table_capture[pid];
    while (p_tmp->next) {
      if (p_tmp->next->signal == sig) {
        p_clr = p_tmp->next;
        p_tmp->next = p_tmp->next->next;
        free(p_clr);
        return 0;
      }
      p_tmp = p_tmp->next;
    }
    return -2; // Le signal n'a pas ete trouve
  }

}

/**
 * Supprime tous les signaux a intercepter de la liste des signaux du processus
 * \param pid : le num du processus dont on veut supprimer un signal
 * \return  rien
 */
void supp_all_sig (int pid) {
  signal_t *p_tmp; 
  signal_t *p_clr; 

  p_tmp = table_capture[pid];
  p_clr = p_tmp;
  while (p_tmp) {
    p_tmp = p_tmp->next;
    free(p_clr);
    p_clr = p_tmp;
  }
}

/**
 * Le code des signaux du language ksinu n'ai pas le meme que les constantes
 * definient dans lib_ksinu.h. Cette fonction permet de traduire un code ksinu en 
 * constante bibliotheque.
 * \param k_sig : le code ksinu a traduire
 * \return  
 * - -1 en cas de signal inexistant
 * - la constante bibliotheque correspondante sinon
 */
int sig_ksinu2lib (int k_sig) {
  int sig;

  switch (k_sig) {
  case 2:
    sig = S_KILL;
    break;
  case 3:
    sig = S_MEM;
    break;
  case 4:
    sig = S_SUSP;
    break;
  case 5:
    sig = S_REP;
    break;
  case 6:
    sig = S_USR1;
    break;
  case 7:
    sig = S_USR2;
    break;
  case 8:
    sig = S_INSTI;
    break;
  default:
    return -1;
  }
  return sig;
}

/**
 * Le code des signaux du language ksinu n'ai pas le meme que les constantes
 * definient dans lib_ksinu.h. Cette fonction permet de traduire une constante 
 * bibliotheque en un code ksinu.
 * \param l_sig : la constante bibliotheque a traduire
 * \return  
 * - -1 en cas de signal inexistant
 * - le code ksinu correspondant sinon
 */
int sig_lib2ksinu (int l_sig) {
  int sig;

  switch (l_sig) {
  case S_KILL:
    sig = 2;
    break;
  case S_MEM:
    sig = 3;
    break;
  case S_SUSP:
    sig = 4;
    break;
  case S_REP:
    sig = 5;
    break;
  case S_USR1:
    sig = 6;
    break;
  case S_USR2:
    sig = 7;
    break;
  case S_INSTI:
    sig = 8;
    break;
  default:
    return -1;
  }
  return sig;
}
