/**
 * @file systeme.c
 * Source de base definissant le fonctionnement du systeme.
 * Gere par exemple :
 * - l'initialisation des differents elements du programme
 * - la reception et l'interpretation des signaux provenant de l'interface
 * S'appuie pour cela sur les fonctions de interpreteur.c 
 */

#include "commandes.h"
#include "systeme.h"

/**************************************************************************************
 Fait par Martin Ludovic
**************************************************************************************/

/**
 * Initialise la memoire principale
 * \return rien
 */
void init_mem () {
  page_t page_mem;
  mot_t mot_mem;
  int i,j;
  int mem_d;

  for (i=0 ; i<4 ; i++) {
    mot_mem[i] = (octet_t)0;
  }
  // Initialisation de la 1ere page
  // Le 1er mot contient le plus petit pid libre
  mot_mem[3] = (octet_t)DEBUT_T_PROC;
  memcpy(page_mem[0], mot_mem, T_MOT);
  // Initialisation de la table des processus
  mot_mem[3] = 0;
  changer_etat(E_MORT,mot_mem);
  for (i=1 ; i<N_MAX_PROC+1 ; i++) {
    memcpy(page_mem[i], mot_mem, T_MOT);
  }
  // Initialisation de la table des pages libres
  mot_mem[0] = (octet_t)2; // le plus petit num de page mem lire
  mot_mem[1] = (octet_t)0; // position courante ds la table des pages swap libres
  mot_mem[2] = (octet_t)PAGE_VIDE;
  mot_mem[3] = (octet_t)PAGE_VIDE;
  for (i=N_MAX_PROC+1 ; i<NMPPP ; i++) {
    if (i == N_MAX_PROC+2) {
      mot_mem[0] = (octet_t)PAGE_VIDE;
      mot_mem[1] = (octet_t)PAGE_VIDE;
    }
    memcpy(page_mem[i], mot_mem, T_MOT);
  }
  memcpy(mem[NUM_PAGE_META], page_mem, T_PAGE);
  // Initialisation de la 2eme page
  for (i=0 ; i<NMPPP ; i++) {
    for (j=0 ; j<T_MOT ; j++) {
      mot_mem[j] = (octet_t)j+i*T_MOT;
    }
    memcpy(page_mem[i], mot_mem, T_MOT);
  }
  memcpy(mem[NUM_PAGE_SWAP], page_mem, T_PAGE);
  // Pour le debuggage...
  mem_d=open("./f.mem",O_CREAT|O_TRUNC|O_RDWR,0666);
  write(mem_d,mem,NPP*NMPPP*T_MOT);
}

/**
 * Initialise la memoire secondaire (swap)
 * \return
 * - -1 en cas d'echec
 * - 0 en cas de reussite
 */
int init_swap () {
  page_t page_swap;
  mot_t mot_swap;
  int i,j;
  int num_page_swap,num_page_suivante;

  for (i=0 ; i<4 ; i++) {
    mot_swap[i] = (octet_t)0;
  }
  // Ouverture du fichier simulant la mem secondaire
  if ((swap_d=open("f.swap",O_RDWR|O_CREAT|O_TRUNC,0666)) == -1) {
    fprintf(stderr,"Impossible d'ouvrir ou de creer le fihier swap\n");
    return -1;
  }
  // Creation d'un swap 'vide'
  // Remplissage du swap pour qu'il fasse NPPS*NMPPP*T_MOT*8 octets
  for (i=0 ; i<NMPPP ; i++) { // Page vide
    memcpy(page_swap[i], mot_swap, T_MOT);
  }
  for (i=0 ; i<NPPS ; i++) {
    if (write(swap_d,page_swap,T_PAGE) != T_PAGE) {
      fprintf(stderr,"Impossible d'initialiser le fihier swap\n");
      return -1;
    }
  }
  if (lseek(swap_d,0,SEEK_SET) == -1) {
    return -1;
  }
  // Creation de la table des pages libres du swap
  num_page_suivante = NMPPP*T_MOT - 1; // Les NMPPP*T_MOT 1eres pages sont stockees ds la mem RAM
  num_page_swap = num_page_suivante + 1;
  while (num_page_swap != NO_MORE_SWAP) {
    // Constitution d'une page contenant des numeros de pages libres
    for (i=0 ; i<NMPPP ; i++) {
      for (j=0 ; j<T_MOT ; j++) {
        if (num_page_swap >= NPPS)
          num_page_swap = NO_MORE_SWAP;
        mot_swap[j] = (octet_t)num_page_swap;
        if (num_page_swap != NO_MORE_SWAP)
          num_page_swap++;
      }
      memcpy(page_swap[i], mot_swap, T_MOT);
    }
    // Ecriture de la page
    if (lseek(swap_d,num_page_suivante*T_PAGE,SEEK_SET) == -1) {
      return -1;
    }
    if (write(swap_d,page_swap,T_PAGE) != T_PAGE) {
      return -1;
    }
    num_page_suivante = num_page_swap-1;
  }
  return 0;
}

/**
 * Lancement d'un programme: creation d'un processus avec une table des pages et remplissage 
 * de la zone de code.
 * Rem:
 * - La zone de pile est allouee dynamiquement durant l'execution du processus
 * - La zone de donnees est fixee par la 1ere instruction (DONNEES)
 * \param fic : nom du fichier contenant le code objet du programme a lancer
 * \return 
 * - -1 en cas d'echec
 * - 0 sinon
 */
int lancer_prog (char* fic, int pid) {
  mot_t mot,mot_tp;
  int num_mot_tp=0;
  page_t page_lue;
  void *page_vide;
  int i;
  int num_table_page,num_page;
  int desc;
  int lu;

  // Cherche un numero de processus libre
  if (pid == 0) {
    if ((pid=pid_libre()) < 1)
      return -1; 
  }
  // Cherche une page libre en memoire principale pour la table des pages du proc
  if ((num_table_page=page_forcee(pid)) < 1) {
    liberer_pid(pid);
    return -1;
  }
  // Initialise cette page
  for (i=0 ; i<T_MOT ; i++) {
    mot[i] = (octet_t)0;
    mot_tp[i] = (octet_t)0;
  }
  for (i=0 ; i<NMPPP ; i++) {
    if (ecrire_mot_mem(i,num_table_page,mot) == -1) {
      liberer_page(num_table_page);
      liberer_pid(pid);
      return -1;
    }
  }
  // Initialise l'entree correspondant au processus ds la table des processus
  changer_type_pp(PP_RAM,mot);
  changer_etat(E_ELIGIBLE,mot);
  mot[3] = (octet_t)num_table_page;
  mot[1] = (octet_t)0;
  mot[0] = (octet_t)S_NULL;
  if (ecrire_mot_mem(pid,NUM_PAGE_META,mot) == -1) {      
    liberer_page(num_table_page);
    liberer_pid(pid);
    return -1;
  }
  // Traitement de la zone de code
  // Ouverture du fichier objet
  if ((desc=open(fic,O_RDONLY)) == -1) {
    liberer_page(num_table_page);
    liberer_pid(pid);
    return -1;
  }
  // Remplissage des pages de code
  if ((page_vide=calloc(1,sizeof(page_t))) == NULL) {
    liberer_page(num_table_page);
    liberer_pid(pid);
    return -1;
  }
  memcpy(page_lue,page_vide,sizeof(page_t));
  if ((lu=read(desc,page_lue,sizeof(page_t))) == -1) {
    liberer_page(num_table_page);
    liberer_pid(pid);
    close(desc);
    free(page_vide);
    return -1;
  }
  if (lu == 0) { // On verifie que le fichier a executer n'est pas vide
    fprintf(stderr,"Le fichier %s est vide!\n",fic);
    liberer_page(num_table_page);
    liberer_pid(pid);
    close(desc);
    free(page_vide);
    return -1;
  }
  while (lu != 0) {
    // Recherche une page physique libre en RAM
    switch (num_page=page_libre(pid)) {
    case -1:
      supprimer_proc(pid); // Supprime le proc et ses pages
      free(page_vide);
      close(desc);
      return -1;
    case -2:
      // Plus de page libre en RAM
      // Recherche une page swap libre
      switch (num_page=page_swap_libre(pid)) {
      case -2:
        fprintf(stderr,"Impossible de continuer le lancement du processus: liberer de la mem\n");
      case -1:
        supprimer_proc(pid);
        free(page_vide);
        close(desc);
        return -1;
      default:
        // Copie la page lue dans une page physique du swap
        if (ecrire_page_swap(page_lue,num_page) == -1) {
          supprimer_proc(pid);
          free(page_vide);
          close(desc);
          return -1;
        }
        // Entre cette page ds la table des pages du processus
        mot_tp[2] = (octet_t)PP_SWAP;
        mot_tp[3] = (octet_t)num_page;
        if (ecrire_mot_mem(num_mot_tp,num_table_page,mot_tp) == -1) {
          supprimer_proc(pid);
          free(page_vide);
          close(desc);
          return -1;
        }
      }
      break;
    default:
      // Copie la page lue dans une page physique de la RAM mot par mot
      for (i=0 ; i<NMPPP ; i++) {
        if (ecrire_mot_mem(i,num_page,page_lue[i]) == -1) {
          supprimer_proc(pid);
          free(page_vide);
          close(desc);
          return -1;
        }
      }
      // Entre cette page ds la table des pages du processus
      mot_tp[2] = (octet_t)PP_RAM;
      mot_tp[3] = (octet_t)num_page;
      if (ecrire_mot_mem(num_mot_tp,num_table_page,mot_tp) == -1) {
        supprimer_proc(pid);
        free(page_vide);
        close(desc);
        return -1;
      }
    }
    num_mot_tp ++;
    memcpy(page_lue,page_vide,sizeof(page_t));
    if ((lu=read(desc,page_lue,sizeof(page_t))) == -1) {
      supprimer_proc(pid);
      free(page_vide);
      close(desc);
      return -1;
    }
  }
  free(page_vide);
  return 0;
}


/**************************************************************************************
 Fait par Ouachani Mourad
**************************************************************************************/

/**
 * Handler interceptant SIGUSR1 qui demande au systeme d'executer une nouvelle instruction
 * \param sig : le num du signal
 * \return rien
 */
void top_horloge_handler (int sig) {
  int pid;
  int i;

  sig = 0; // juste pour eviter au compilo de raler...
  // Interprete tous les signaux de la table des processus
  for (i=1 ; i<=N_MAX_PROC ; i++) {
    interprete_signaux(pid);
  }
  // Recherche le prochain processus a executer
  pid = liste_next_pid(liste_proc);
  // interprete l'instruction courante du processus pid
  next_inst(pid);
}

/**
 * Handler interceptant SIGUSR2 qui averti le systeme qu'une requete a ete envoyee ds le tube
 * \param sig : le num du signal
 * \return rien
 */
void interface_handler (int sig) {
  sig=0;
  lire_tube();
}

/**
 * Lire le contenue du tube
 * \return 
 * - -1 en cas d'echec
 * - 0 sinon
 */
int lire_tube(){
  struct message mes_lus;
  int lu;

  if( (fifo = open( FIFO_NAME, O_RDONLY|O_NONBLOCK ))==-1 ){
    printf("erreur d'ouverture du tube\n");
    return -1;
  } 
  lu = 1;
  while (lu > 0) {
    lu = read(fifo,&mes_lus,sizeof(struct message));
    if ((lu != sizeof(struct message)) && (lu != 0)) {
      printf("erreur de lecture dans le tube\n");
      return -1;
    }
  }
  close(fifo);
  interprete_commande(mes_lus);
  return 0;
}

/**
 * Interpreter les commandes envoyes par l'interface
 * \param structure message  contenant la commande ainsi que les arguments
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 **/
int  interprete_commande(message mes){
  switch(mes.commande){
  case LP:
    liste_process();
    break;
  case STDIN:
    modif_process(mes.int_opt,mes.ref);
    break;
  case LPT:
    printtp(mes.int_opt);
    break;
  case PPP:
    printpp(mes.int_opt);
    break;
  case PSIG:
    printsig(mes.int_opt);
    break;
  case SIG: 
    signal_proc(mes.int_opt2,mes.int_opt);
    break;
  default:
    perror("erreur de message\n");
    return -1;
  }
  return 0;
}

/**
 * Ecrit une chaine de caractere dans le tampon d'entree dun processus
 * \param numero du processus
 * \param chaine de carracteres
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 **/ 
int modif_process(int num_proc, char *ref){
  int i=0;

  while(ref[i] != '\0'){
    if(ecrire_stdin(num_proc, ref[i]) < 0){
      printf("impossible d'ecrire dans le tampon d'entree du processus N : %d\n",num_proc);
    }
    i++;
  }
  return 0;
}

/**
 * Lister tous les processus actif
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 **/
int liste_process(){
  int i;
  mot_t mot;
  int etat;
  
  printf("PID\n");
  for(i=1; i<N_MAX_PROC+1;i++){
    if(lire_mot_mem(i, NUM_PAGE_META, mot) == -1){
      printf("erreur de lecture du mot du processus N %d\n",i);
      return -1;
    }
    etat = lire_etat(mot);
    switch(etat){
    case E_MORT:
      printf("Processus N %d mort\n", i);
      break;
    case E_ELIGIBLE: 
      printf("Processus N %d eligible\n",i);
      break;
    case E_ZOMBI :    
      printf("Processus N %d zombi\n",i);
      break;
    case E_SUSP:     
      printf("Processus N %d suspendu\n",i);
      break;
    case E_ATT_FILS:
      printf("Processus N %d attendant la mort d'un fils\n",i);
      break;
    case E_ATT_LECT: 
      printf("Processus N %d attendant une fin de lecture\n",i);
      break;
    case E_ATT_ECRIT: 
      printf("Processus N %d attendant une fin d'ecriture\n",i);
      break;
    default:
      printf("erreur etat non attendu\n");
      return -1;      
    }
  }
  return 0;
}

/**
 * Afficher la table des pages d'un processus
 * \param numero de processus
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 **/
int printtp(pid_t pid){
  mot_t mot;
  int num_page_tp;
  int i,j;
  char *tab;

  if(lire_mot_mem(pid, NUM_PAGE_META, mot) == -1){
    printf("Erreur de lecture du mot dans la page physique numero %d\n", pid);
    return -1;
  }
  tab = malloc(1);
  num_page_tp = (int)mot[3];
  for(i=0; i<3;i++){
    for(j=i*T_MAX_ZONES;j<(i+1)*T_MAX_ZONES;j++){
      if(lire_mot_mem(i,num_page_tp,mot) == -1){
        printf("Erreur de lecture du mot dans la page physique numero %d\n", j);
        return -1;
      } 
      if ((int)mot[2] == PP_RAM) {
        tab = realloc(tab,4);
        memcpy(tab ,"RAM",4);
      }
      else if ((int)mot[2] == PP_SWAP) {
        tab = realloc(tab,5);
        memcpy(tab,"SWAP",5);
      }
      else if ((int)mot[2] == PP_NULL) {
        break;
      }
      switch(i){
      case 0:
        printf("%d e  page de code :  %d en %s\n",j+1,(int)mot[3],tab);
        break;
      case 1:
        printf("%d e  page de donnee :  %d en %s\n",j+1,(int)mot[3],tab);
        break;
      case 2:
        printf("%d e  page de pile :  %d en %s\n",j+1,(int)mot[3],tab);
        break;
      default:
        printf("page inconnue\n");
        return -1;
      }
      printf("\n");
    }
  }
  free(tab);
  return 0;
}

/**
 * Ecris dans un fichier le contenue d'une page physique
 * \param numero de la page
 * \return
 * - -1 en cas d'erreur
 * - 0 sinon
 **/
int printpp(int num){
  int i;
  int d;
  mot_t mot;
  if((d=open("table.mem",O_WRONLY|O_CREAT)) == -1){
    printf("erreur d'ouverture du fichier ou de creation\n");
    return -1;
  }
  for(i = 0; i < NMPPP; i++){
    if(lire_mot_mem(i,num, mot) == -1){
      printf("erreur de lecture du mot\n");
      return -1;
    }
    if(write(d,mot,4) == -1){
      printf("erreur d'ecriture dans le Fichier\n");
      return -1;
    }
    i++;
  }
  close(d);
  return 0;
}

/**
 * Fonction qui affiche les informations relatives aux signaux
 * \param pid : num du processus
 * \return  
 * - -1 en cas d'erreur inconnue
 * - 0  sinon
 */
int printsig(int num_proc){
  mot_t mot;
  if(lire_mot_mem(num_proc,NUM_PAGE_META,mot) == -1){
    printf("erreur de lecture du mot N de process : %d\n",num_proc);
    return -1;
  }
  if(SIFUNSIG(mot[0])){
    printf("Aucun signal attache au processus N %d\n", num_proc);
    return 0;
  }
  else {
    if( SIFKILL(mot[0])){
      printf("Signal: Arme Fatale attache au processus N %d\n", num_proc);
    }
    if( SIFMEM(mot[0])){
      printf("Signal: violation de memoire  attache au processus N : %d\n", num_proc);
    }
    if( SIFSUSP(mot[0])){
      printf("Signal: suspension attache au processus N %d\n", num_proc);
    }
    if( SIFREP(mot[0])){
      printf("Signal: de reprise attache au processus N %d\n", num_proc);
    }
    if( SIFUSR1(mot[0])){
      printf("Signal: non definie attache au processus N %d\n", num_proc);
    }
    if( SIFUSR2(mot[0])){
      printf("Signal: non defenie attache au processus N %d\n", num_proc);
    }
    if( SIFINSTI(mot[0])){
      printf("Signal:  instruction illegale  attache au processus N %d\n", num_proc);
    }
  }
  return 0;  
}

/**
 * Fonction qui affecte un signale a un processus
 * \param num_sig : num du signal
 * \param pid : num du processus
 * \return  
 * - -1 en cas d'erreur inconnue
 * - 0  sinon
 */
int signal_proc(int num_sig, int pid){
  mot_t mot;
  if(lire_mot_mem(pid,NUM_PAGE_META,mot) == -1){
    printf("echec a la lecture du mot , processus N%d\n", pid);
    return -1;
  }
  mot[0] = mot[0] | num_sig ;
  if(ecrire_mot_mem(pid,NUM_PAGE_META,mot) == -1){
    printf("echec de l'ecriture du mot , N processus :%d\n",pid);
    return -1;
  }
  return 0;
}
